# -*- coding: utf-8 -*-
from six.moves import range
import numpy as np



def diffarray(x, r=None, include=True, aniso=False, gui_args=()):
    '''
    Create the unique pair index across the coordinates of x

    Syntax: index_head,index_tail,diff = diffarray(x,r=None,include=True)

    Input:    

    x           n by nd     2D array    row for data and col for dimension
    r           scalar                  the range to be considered
    include     bool                    True denotes diff include self
    aniso       bool                    True consider anisotropy 
                                        (x.shape[1] must be 2)
    gui_args                tuple       a tuple contains gui elements
    Output:

    index_head  n           1D array    the first index of the pairs
    index_tail  n           1D array    the second index of the pairs
    diff        n by nd     2D array    the differerces between the pairs at
                                          each dimension
    angdiff     n           1D array    the angles in radius in which the
                                        horizontal x-direction are considered to
                                        be zero degrees
    '''

    row_count, col_count = x.shape
    d = 0 if include == True else 1
    if col_count != 2 and aniso == True: # error 
        raise ValueError("aniso is True but x's dimension is not 2")
    k = 0
    nn0 = (row_count - d)*(row_count - d + 1)/2
    hasMemoryError = False
    try:
        index_head = np.zeros(nn0)
        index_tail = np.zeros(nn0)
        diff = np.zeros((nn0, col_count))
        angdiff = np.zeros(nn0) if aniso else None
    except MemoryError, memerr:
        hasMemoryError = True

    if r is None:
        if hasMemoryError:
            raise memerr # we have no solution here...
        else:
            for i in range(row_count - d):
                index_head[k:k+row_count-i-d] = i
                index_tail[k:k+row_count-i-d] = np.arange(row_count)[i+d:]
                diff[k:k+row_count-i-d]=x[i]-x[np.arange(row_count)[i+d:]]
                if aniso is True:
                    angdiff[k:k+row_count-i-d] = \
                      np.arctan(
                          diff[k:k+row_count-i-d][:,1]
                          / diff[k:k+row_count-i-d][:,0]
                          )
                # if symmetric is False:
                #     Comment out because the pairs only consider
                #     the paris of upper triangle
                #     angdiff[k:k+row_count-i-d][diff[k:k+row_count-i-d][:,0]<0]+=np.pi
                k = k + row_count - i - d
            index_head = index_head.astype(np.int)
            index_tail = index_tail.astype(np.int)
    else: # consider r
        if hasMemoryError: # try split x
            print 'Try to split x...'
            index_head = np.array([], dtype=np.int64)
            index_tail = np.array([], dtype=np.int64)
            diff = np.array([]).reshape((0, col_count))
            angdiff = np.array([]) if aniso else None
            step_len = 100
            step_count = int(np.ceil(float(row_count) / step_len))
            current_count = 0
            if gui_args:
                title = DataObj.getProgressText()
                gui_args[0].setProgressRange(0, step_count)
                sub_title = title + "\n- Diffarray..."
                gui_args[0].setCurrentProgress(0, sub_title)
            for step in range(0, row_count, step_len):
                piece_x = x[step:step+step_len, :]
                for step2 in range(step, row_count, step_len):
                    piece_x2 = x[step2:step2+step_len, :]
                    if step == step2: #need upper-triangle
                        idx_head_step, idx_tail_step, diff_step, angdiff_step =\
                            diffarray(piece_x, r, include, aniso)
                    else: #need square
                        idx_head_step, idx_tail_step, diff_step, angdiff_step =\
                            cdist(piece_x, piece_x2, r, aniso)
                    idx_head_step += step
                    idx_tail_step += step2
                    index_head = np.append(index_head, idx_head_step)
                    index_tail = np.append(index_tail, idx_tail_step)
                    diff = np.append(diff, diff_step, axis=0)
                    angdiff = np.append(angdiff, angdiff_step) if aniso else None
                    current_count+=1
                    if gui_args and not gui_args[0].wasProgressCanceled():
                        gui_args[0].setCurrentProgress(
                            current_step_count ,
                            sub_title + '({c}/{n})'.format(
                                c=current_count,
                                n=step_count
                                )
                            ) #rest
                        gui_args[0].drawGUI()
                    print str(current_count)+ '/' +str(step_count*(step_count+1)/2)
                    print 'size:', index_head.shape
        else:
            nn = nn0
            for i in range(row_count):
                index_head_=np.zeros(row_count-i-d)
                index_tail_=np.zeros(row_count-i-d)
                diff_=np.zeros((row_count-i-d,col_count))
                index_head_[:]=i
                index_tail_[:]=np.arange(row_count)[i+d:]
                diff_=x[i]-x[np.arange(row_count)[i+d:]]
                if aniso is True:
                    angdiff_ = np.arctan2(diff_[:,1], diff_[:,0])
                    # if symmetric is False:
                    #   angdiff_[diff_[0]<0]+=np.pi
                else:
                    angdiff_ = None
                idxx = np.where(
                    np.sqrt((diff_**2).sum(axis=1, keepdims=True))<=r
                    )[0]
                ni = idxx.size
                index_head[k:k+ni] = index_head_[idxx]
                index_tail[k:k+ni] = index_tail_[idxx]
                diff[k:k+ni,:] = diff_[idxx]
                if aniso is True:
                    angdiff[k:k+ni] = angdiff_[idxx]
                k = k + ni
                if nn-k <= row_count-i:
                    index_head = np.append(index_head, np.empty(nn0))
                    index_tail = np.append(index_tail, np.empty(nn0))        
                    diff=np.vstack((diff,np.empty((nn0, col_count))))
                    if aniso is True:
                        angdiff = np.append(angdiff, np.empty(nn0))
                    nn = nn + nn0

            index_head=index_head[:k].astype(np.int)
            index_tail=index_tail[:k].astype(np.int)
            diff=diff[:k]
            if aniso:
                angdiff=angdiff[:k]
    
    return index_head, index_tail, diff, angdiff

def diffarray_split(x, r=None,
    include=True, aniso=False,
    split_length=3000, container_size_step=1000000, show_info=False,
    gui_args=()):
    '''solve memory problem'''

    def _enlarge_container(
        index_head, index_tail, diff, angdiff,
        maximum_container_size, container_size_step):
        if show_info:
            print 'enlarge containter'
        index_head = np.append(
            index_head, np.empty(
                container_size_step, dtype=np.int64
                )
            )
        index_tail = np.append(
            index_tail, np.empty(
                container_size_step, dtype=np.int64
                )
            )        
        diff = np.vstack(
            (diff, np.empty((container_size_step, col_count)))
            )
        if aniso is True:
            angdiff = np.append(
                angdiff, np.empty(
                    container_size_step, dtype=np.int64
                    )
                )
        maximum_container_size += container_size_step
        return (index_head, index_tail, diff, angdiff,
            maximum_container_size)

    row_count, col_count = x.shape
    d = 0 if include == True else 1
    if col_count != 2 and aniso == True: # error 
        raise ValueError("aniso is True but x's dimension is not 2")

    index_head = np.empty(container_size_step, dtype=np.int64)
    index_tail = np.empty(container_size_step, dtype=np.int64)
    diff = np.empty((container_size_step, col_count))
    angdiff = np.empty(container_size_step) if aniso else None
    step_len = split_length
    step_count = int(np.ceil(float(row_count) / step_len))
    total_step = (step_count+1) * step_count / 2
    current_count = 0
    current_data_size = 0
    maximum_container_size = container_size_step

    if gui_args:
        title = gui_args[0].getProgressText()
        gui_args[0].setProgressRange(0, total_step)
        sub_title = '\n'.join((title, gui_args[1]))
        gui_args[0].setCurrentProgress(0, sub_title)

    for step in range(0, row_count, step_len):
        piece_x = x[step:step+step_len, :]
        for step2 in range(step, row_count, step_len):
            piece_x2 = x[step2:step2+step_len, :]
            if step == step2: #need upper-triangle
                idx_head_step, idx_tail_step, diff_step, angdiff_step =\
                    diffarray(piece_x, r, include, aniso)
            else: #need square
                idx_head_step, idx_tail_step, diff_step, angdiff_step =\
                    cdist(piece_x, piece_x2, r, aniso)
            idx_head_step += step
            idx_tail_step += step2

            data_size_step = idx_head_step.size
            current_data_size += data_size_step
            while current_data_size > maximum_container_size:
                (index_head, index_tail, diff, angdiff,
                maximum_container_size) =\
                    _enlarge_container(
                        index_head, index_tail, diff, angdiff,
                        maximum_container_size, container_size_step)
            slice_idx = slice(
                current_data_size-data_size_step, current_data_size)
            index_head[slice_idx] = idx_head_step
            index_tail[slice_idx] = idx_tail_step
            diff[slice_idx, :] = diff_step
            if aniso:
                angdiff[slice_idx] = angdiff_step

            # index_head = np.append(index_head, idx_head_step)
            # index_tail = np.append(index_tail, idx_tail_step)
            # diff = np.append(diff, diff_step, axis=0)
            # angdiff = np.append(angdiff, angdiff_step) if aniso else None
            current_count+=1
            if show_info:
                print str(current_count)+ '/' +str(step_count*(step_count+1)/2)
            if gui_args:
                if not gui_args[0].wasProgressCanceled():
                    gui_args[0].setCurrentProgress(
                        current_count ,
                        sub_title + '({c}/{n})'.format(
                            c=current_count,
                            n=total_step
                            )
                        ) #rest
                    gui_args[0].drawGUI()
                else:
                    return False
    if gui_args:
        gui_args[0].setCurrentProgress(text=title)
    slice_final = slice(0, current_data_size)
    index_head = index_head[slice_final]
    index_tail = index_tail[slice_final]
    diff = diff[slice_final, :]
    if aniso:
        angdiff = angdiff[slice_final]

    return index_head, index_tail, diff, angdiff

def multiarray(x,include=True):
    '''
    like diffarray but operator is multilize
    '''
    lenx=len(x)
    diff=[]

    index_head=[]
    index_tail=[]
        
    if include == True:
        for i in range(lenx):
            index_head+=[i]*(lenx-i)
            index_tail+=range(lenx)[i:]
        for i in range(lenx):  
            for j in range(i,lenx):
                #index_head.append(i)
                #index_tail.append(j)
                diff.append(x[i]*x[j])
    elif include == False:
        for i in range(lenx-1):
            index_head+=[i]*(lenx-i-1)
            index_tail+=range(lenx)[i+1:]
        for i in range(lenx): 
            for j in range(i+1,lenx):
                #index_head.append(i)
                #index_tail.append(j)
                diff.append(x[i]*x[j])
    
    index_head,index_tail,diff=map(np.array,(index_head,index_tail,diff))
    return index_head,index_tail,diff

def cdist(x, y, r, aniso):
    x_row, x_col = x.shape
    y_row, y_col = y.shape
    if x_col != y_col:
        raise ValueError(
            'two array must have the same column length')
    if x_col != 2 and aniso == True:
        raise ValueError(
            'anisotropy is only suit for column length equal to 2')

    index_head = np.empty( x_row * y_row, dtype=np.int64)
    index_tail = np.empty( x_row * y_row, dtype=np.int64)    
    for i in range(x_row):
        index_head[i*y_row:(i+1)*y_row] = i
        index_tail[i*y_row:(i+1)*y_row] = np.arange(y_row)

    diff = np.empty((x_row*y_row, x_col))
    for i in range(x_col):
        diff[:,i] = (x[:,i:i+1] - y[:,i]).ravel()
    if aniso == True:
        angdiff = np.arctan2(diff[:,1], diff[:,0]).ravel()
    else:
        angdiff = np.empty(x_row*y_row)

    if r != -1.0:
        dis_arr = np.zeros(x_row*y_row)
        for i in range(x_col):
            dis_arr = dis_arr+diff[:,i]**2

        idxx = np.where(np.sqrt(dis_arr) <= r)
        index_head = index_head[idxx]
        index_tail = index_tail[idxx]
        diff = diff[idxx]
        if aniso == True:
            angdiff = angdiff[idxx]

    return index_head.ravel(), index_tail.ravel(), diff, angdiff.ravel()

if __name__ == "__main__":
    import time

    x = np.array(np.random.random((10,2)))
    r = 0.5
    include = True
    aniso = True
    stime=time.time()
    h1, t1, d1, a1 = diffarray(x, r, include, aniso)
    print 'No split time cost:', time.time() - stime
    stime=time.time()
    h1s, t1s, d1s, a1s = diffarray_split(x, r, include, aniso,show_info=True)
    print 'Split time cost:', time.time() - stime



import os
import platform
import struct
import shutil
import subprocess
from numba.pycc import CC

FILE_FOLDER = os.path.dirname(__file__)
if platform.platform().startswith('Darwin'):
    OS = 'Mac'
elif platform.platform().startswith('Windows'):
    OS = 'Windows'

ARCHITECTURE = 8 * struct.calcsize("P") # 32 or 64 bit 
if ARCHITECTURE == 64:
    ARCHITECTURE = 'x64'
elif ARCHITECTURE == 32:
    ARCHITECTURE = 'x86'
else:
    raise ValueError('No Architecture Supported.')

ROOT_FOLDER = os.path.join(FILE_FOLDER, OS, ARCHITECTURE)


if __name__ == "__main__":
    package_path = os.path.abspath(os.path.dirname(__file__))
    compile_files_py = []
    for dd, sub_dd, ff in os.walk( package_path ):
        compile_files_py += [(dd,i) for i in ff if i.endswith('__aot.py')]
    for dd, ff in compile_files_py:
        subprocess.call(['python', os.path.join(dd,ff)])
        module_path = dd
        module_name = ff[:-8]

        compiled_files = []
        for f in os.listdir(module_path):
            if f.startswith(module_name) and f.endswith(('.so', '.pyd')):
                compiled_files.append(f)
    
        path_p_to_m = os.path.relpath(module_path, package_path)

        for f in compiled_files:
            shutil.move(
                os.path.join(module_path,f),
                os.path.join(package_path, 'aot', path_p_to_m, f)
                )

    # add __init__.py if needed
    for d, sub_d, fs in os.walk(os.path.join(package_path,'aot')):
        if '__init__.py' not in fs: # need
            finit = open(os.path.join(d,'__init__.py'),'w')
            finit.close()



# -*- coding: utf-8 -*-
import numpy

from .regression import regression
from .designmatrix import designmatrix


def localmeanBME( ck, ch, cs, zh, ms, vs,
                  Khh, Ksh, Kss, order ):

    nh = zh.shape[0]
    ns = ms.shape[0]
    mkest = 0.
    mhest = numpy.empty((nh, 1))*numpy.NaN
    msest = numpy.empty((ns, 1))*numpy.NaN
    vkest = 0.

    c = numpy.vstack((ch, cs))
    ms=ms.reshape((ns,1)) 
    vs=vs.reshape((ns,1))       
    Ksh=Ksh.reshape((ns,nh))
    Kss=Kss.reshape((ns,ns))
    #add '_' b/s we won't to change vs and Kss inplace
#    vs_ = numpy.tile( vs,(1,ns ) ) if len(vs) else vs
#    Kss_ = Kss + numpy.diag(vs_)
    K = numpy.vstack( (numpy.hstack((Khh, Ksh.T)), numpy.hstack((Ksh, Kss)) ) )
    z = numpy.vstack((zh, ms))

    best, Vbest, mm, index = regression(c, z, order, K)

    if mm.size > 0:
        mhest = mm[:nh,0:1]
        msest = mm[nh:,0:1]
    else:
        mhest[:] = 0.0 #numpy.array([])
        msest[:] = 0.0 #numpy.array([])

    x, index = designmatrix(ck, order)
    if x.size > 0:
        mkest = ( x.dot(best) )[0][0]
        vkest = ( x.dot(Vbest).dot(x.T) )[0][0]

    return mkest , mhest, msest, vkest



# -*- coding: utf-8 -*-
"""
Created on Fri Jul  3 08:48:06 2015

@author: hdragon689
"""
import numpy as np
import shapefile as shp
import matplotlib.pyplot as plt
#  import matplotlib.patches as patches
from matplotlib.patches import Polygon
#  from matplotlib.collections import PatchCollection


def polygonbkplot(shpdata,ax=None,maxx=None,minx=None,maxy=None,miny=None):
  '''
  Plot polygon shapefile without filling and can be used as a background of 
  other maps
  
  Syntax: 
    polygonbkplot(shpdata,ax=None,maxx=None,minx=None,maxy=None,miny=None)
    
  Input: 
    shpfile   string      the path of shapefile
    ax        ax          optional. the axes object of the map to be plotted. 
                          Details of the axes can refer to matplotlib package
    maxx      scalar      optional. maxx, minx, maxy, miny are coordinates of 
    minx                  longitude and latitude, respectively, that define the  
    maxy                  rectangular boundary of the map in the ax
    miny                      
                          
  '''  
  
  sf = shp.Reader(shpdata)
  recs    = sf.records()
  shapes  = sf.shapes()
  Nshp    = len(shapes)
  # cns     = []
  # for nshp in xrange(Nshp):
  #   cns.append(recs[nshp][1])

  # cns = np.array(cns)
#  cm    = plt.get_cmap('Dark2')
#  cccol = cm(1.*np.arange(Nshp)/Nshp)
  if ax is None:
    ax=plt.figure().add_subplot(111)

  for nshp in xrange(Nshp):
#    ptchs   = []
    pts     = np.array(shapes[nshp].points)
    prt     = shapes[nshp].parts
    par     = list(prt) + [pts.shape[0]]
    for pij in xrange(len(prt)):
      ax.add_patch(Polygon(pts[par[pij]:par[pij+1]],fill=False,closed=True))     
  
  if [maxx,minx,maxy,miny]!=[None]*4:    
    ax.set_xlim(minx,maxx)
    ax.set_ylim(miny,maxy) 


# -*- coding: utf-8 -*-
import numpy as np
import pandas as pd
from scipy.spatial import cKDTree
from scipy.spatial.distance import cdist as scipy_cdist

from ..general.neighbours import neighbours_index_kd
from ..stest.idw import idw_est_coord_value


# need fix this function #
def idw_st(c, z, c_est, n_max=100,
    s_range=None, t_range=None, st_ratio=None, power=2):

    def check_arg(c, n_max, s_range, t_range, st_ratio):
        def get_max_distance(c):
            from np import nanmax
            from scipy.spatial.distance import pdist
            print 'cal pdist...'
            d = pdist(c)
            return nanmax(d)
        n_max = 100 if n_max is 100 else int(n_max)
        s_range = get_max_distance(c[:,:-1]) if s_range is None else s_range
        t_range = get_max_distance(c[:,-1:]) if t_range is None else t_range
        st_ratio = s_range / t_range if st_ratio is None else st_ratio
        return n_max, s_range, t_range, st_ratio

    def make_tree(c, st_ratio):
        c_norm = np.copy(c)
        c_norm[:, -1] = c_norm[:, -1] * st_ratio
        c_tree = cKDTree(c_norm)
        return c_norm, c_tree

    n_max, s_range, t_range, st_ratio =\
        check_arg(c, n_max, s_range, t_range, st_ratio)
    print 'make c tree...'
    c_norm, c_tree = make_tree(c, st_ratio)
    print 'make c est tree...'
    c_est_norm, c_est_tree = make_tree(c_est, st_ratio)
    d_max_norm = (s_range**2 + (t_range * st_ratio)**2)**0.5

    z_est = np.empty((c_est.shape[0],1))
    est_list = [] # c_idx, est_idx

    print 'start neighbouring...'
    # try:
    #     est_dict = unpickleDictionary('est_dict.cpkl')
    # except Exception, e:
    #     print e
    est_dict = neighbours_index_kd(c_est_norm, c_tree, n_max, d_max_norm)
    #     pickleDictionary(est_dict, 'est_dict.cpkl')
    # import pdb
    # pdb.set_trace()
    
    print 'start idw...'
    print c_est.shape 
    cc = 0
    old_cc = 0
    for c_idx, est_idx in est_dict.iteritems():
        est_idx = np.array(est_idx, dtype=int)
        c_idx = np.array(c_idx, dtype=int)

        picked_c = c[c_idx, :]
        picked_z = z[c_idx, :]

        if len(c_idx) == 0:
            print 'no neighbors... set it to np.nan.'
            z_est[c_idx,:] = np.nan
            est_count = est_idx.shape[0]
            cc+=est_count
            if cc - old_cc >= 2000:
                print cc ,'/', c_est.shape
                old_cc = cc
            continue
            # import pdb
            # pdb.set_trace()
        
        est_count = est_idx.shape[0]
        cc+=est_count
        if cc - old_cc >= 2000:
            print cc ,'/', c_est.shape
            old_cc = cc

        split_count = np.ceil(est_count/250.)
        for est_idx_piece in np.array_split(est_idx, split_count):
            picked_c_est = c_est[est_idx_piece, :]
            picked_z_est =\
                idw_est_coord_value(picked_c, picked_z, picked_c_est, power = 2)
            if pd.isnull(picked_z_est).any():
                print 'est has nan result... stop.'
                import pdb
                pdb.set_trace()

            z_est[est_idx_piece, :] = picked_z_est
    print cc ,'/', c_est.shape
    return z_est

# need fix this function #
def outlier_detection(
    data_frame,
    id_column=None,
    spatial_column=None,
    temporal_column=None,
    value_column=None,
    normalized=True,
    time_group_frequency=None,
    search_parameters={
        'bandwith_start': 1.5,
        'bandwith_step': 0.1,
        'bandwith_end': 2,
        'max_count': 100
    },
    inverse_distance_weighted_parameters=None):
    '''
    data_frame: pd dataframe
    coords: lon, lat, time
    values: values for each column
    normalized: zscore at each id
    '''
    data_frame = data_frame.dropna(subset=value_column)
    id_coord = data_frame[[id_column]+spatial_column]\
        .drop_duplicates()\
        .set_index(id_column)

    if normalized: #need to get mean, std at each id(location)
        grp = data_frame.groupby(id_column)
        agg_dict = {}
        for vc in value_column:
            agg_dict[vc] = [np.mean, np.std]
        id_zscore = grp.agg(agg_dict)
        id_zscore.columns = id_zscore.columns.map(('{0[0]}_{0[1]}'.format))
        for vc in value_column:
            id_zscore.loc[id_zscore[vc+'_std'].isnull(),[vc+'_std']] = 1. #trick one point
            id_zscore.loc[id_zscore[vc+'_std']==0,[vc+'_std']] = 1. #trick 0 std

    data_frame =\
        pd.merge(
            data_frame,
            id_zscore,
            how='inner', left_on=id_column, right_index=True,
            sort=True, copy=True
            ) #use inner to check matched length

    data_frame['time_int'] =\
        (data_frame[temporal_column] - data_frame[temporal_column].min())\
        / np.timedelta64(1,'s')

    for vc in value_column:
        #zscore data
        data_frame[vc+'_zs'] =\
            (data_frame[vc] - data_frame[vc+"_mean"]) / data_frame[vc+"_std"]
        #setting outlier flag
        data_frame['outlier_'+vc+'_idw'] = False
        data_frame[vc+'_zs_idw'] = np.nan

    #do idw, but only use past data
    gp_time = data_frame.groupby(
        pd.Grouper(key=temporal_column, freq=time_group_frequency,
            )
        )
    gp_len = len(gp_time)
    for idx, (name, gp_df) in enumerate(gp_time):
        # if idx >= 2500:
        #     break
        if len(gp_df) > 0: #has data in group
            for vc in value_column:
                filtered_data = []
                orig_bs = search_parameters['bandwith_start']
                while (len(filtered_data) < search_parameters['max_count'])\
                    and (search_parameters['bandwith_start'] <= search_parameters['bandwith_end']):

                    search_parameters['bandwith_start'] += search_parameters['bandwith_step']
                    filtered_data =\
                        data_frame[
                            (data_frame[temporal_column] < gp_df[temporal_column].min())
                            & (data_frame[temporal_column] >
                               gp_df[temporal_column].min() - pd.Timedelta(search_parameters['bandwith_start'], 'h'))
                        ] #get filtered data
                    filtered_data = filtered_data[~filtered_data['outlier_'+vc+'_idw']] #remove outlier at vc
                if filtered_data.size == 0: # no valid data
                    data_frame.loc[gp_df.index, vc+"_zs_idw"] = np.nan
                else:
                    filtered_data_2 = filtered_data
                    # filtered_data_2 = pd.merge(
                    #     filtered_data, id_coord,
                    #     how='inner', left_on=id_column, right_index=True,
                    #     sort=True, copy=True
                    #     )
                    gp_df_2 = gp_df
                    # gp_df_2 = pd.merge(
                    #     gp_df, id_coord,
                    #     how='inner', left_on=id_column, right_index=True,
                    #     sort=True, copy=True
                    #     )

                    # if idx == 11:
                    #     import pdb
                    #     pdb.set_trace()
                    rr = idw_st(
                        filtered_data_2[spatial_column + ['time_int']].values,
                        filtered_data_2[[vc+"_zs"]].values,
                        gp_df_2[spatial_column + ['time_int']].values, n_max=100,
                        s_range=10000, t_range=2*3600, st_ratio=10000/3600., power=2) # 10km/1hr
                    data_frame.loc[gp_df.index, vc+"_zs_idw"] = rr
                    outlier_bool =\
                        (data_frame.loc[gp_df.index, vc+"_zs"] - data_frame.loc[gp_df.index, vc+"_zs_idw"]).abs() > 3
                    data_frame.loc[outlier_bool.index[outlier_bool],'outlier_'+vc+'_idw'] = True
                print '{f} -- {a}/{b}, data count: {c}, hour: {d}'.format(
                    f=vc, a=idx+1, b=gp_len,
                    c=filtered_data.shape[0], d=search_parameters['bandwith_start'])
                search_parameters['bandwith_start'] = orig_bs
        else:
            continue
    return data_frame

def time_series_outlier_detection(
    time_series, estimation_datetime_range=None, group_frequency=None,
    standard_deviation_mutiplier=3,
    maximum_outlier_duration=None,
    maximum_outlier_count=0,
    search_parameters={
        'bandwith':
            {'start': '6 hours',
             'step': '6 hours',
             'end': '6 hours'},
        'max_count': 100
        },
    inverse_distance_weighted_parameters={
        'power': 1.5
        },
    print_step=2000):

    '''
    outlier detection for time series

    time_series:
        pandas time series
    estimation_datetime_range: 
        a list of estimation datetime range, e.x. ['2016-03-04', '2017-03-01']
        if None, use time_series
    group_frequency:
        pandas timedelta obejct or any string can be convert to
        used for time series to groupby
        if not use, set it to the data's minimum time unit, like '1 s'
        it is very useful for large dataset with larger group frequency,
        it will give a fast result at a glance
        e.x. '30 minites'
    standard_deviation_mutiplier:
        a number,
        threshold of determind whether data is outlier or not.
    maximum_outlier_duration:
        pandas timedelta obejct or any string can be convert to
        it used for correct the detection, if all outlier are continually
        exists for a duration larger than maximum outlier duration,
        it will set all these outliers to False
        (not outlier, e.g. detection error)
    search_parameters:
        e.x. search_parameters={
            'bandwith':
                {'start': '6 hours',
                 'step': '1 hours',
                 'end': '12 hours'},
            'max_count': 100
            }
        to do...
    inverse_distance_weighted_parameters:
        a dictionary for idw function
    '''

    def __convert_param(search_parameters):
        search_parameters['bandwith']['start'] =\
            pd.Timedelta(search_parameters['bandwith']['start'])
        search_parameters['bandwith']['step'] =\
            pd.Timedelta(search_parameters['bandwith']['step'])
        search_parameters['bandwith']['end'] =\
            pd.Timedelta(search_parameters['bandwith']['end'])
        return search_parameters

    #get estimation datetime
    if estimation_datetime_range:
        estimation_time_series =\
            time_series[
                estimation_datetime_range[0]:estimation_datetime_range[1]]
    else:
        estimation_time_series = time_series


    outlier = pd.Series(0, estimation_time_series.index, name='outlier')
    maximum_outlier_duration = pd.Timedelta(maximum_outlier_duration)
    search_parameters = __convert_param(search_parameters)

    gp = estimation_time_series.groupby(pd.Grouper(freq=group_frequency))
    gp_len = len(gp)
    for idx, (i, ss) in enumerate(gp):
        if len(ss) > 0: #has data in time slice

            #outlier detection
            current_bandwith = search_parameters['bandwith']['start']
            picked_data = time_series[
                    (time_series.index >= i - current_bandwith)\
                        & (time_series.index < i + pd.Timedelta(group_frequency))]

            #get outlier in picked data, remove it
            outlier_in_picked =\
                 outlier[outlier.index.isin(picked_data.index)]
            outlier_in_picked_index =\
                outlier_in_picked[outlier_in_picked == 1].index
            picked_data = picked_data.drop(index=outlier_in_picked_index)

            value = picked_data.values.reshape((-1,1))
            value_mean = value.mean()
            value_std = value.std()
            #avg mean
            outlier_bool =\
                np.abs(ss.values - value_mean)\
                > (standard_deviation_mutiplier * value_std)
            if outlier_bool.any():
                try:
                    for ii in ss[outlier_bool].index:
                        outlier.loc[ii] = 1
                except AttributeError, e:
                    print e
                    import pdb
                    pdb.set_trace()
            else:
                pass

            #event detection
            # current_bandwith = search_parameters['bandwith']['start']
            # picked_data = time_series[
            #         (time_series.index >= i - current_bandwith)\
            #             & (time_series.index < i)]
            # outlier_in_picked =\
            #     outlier[outlier.index.isin(picked_data.index)]

            # if maximum_outlier_duration:
            #     outlier_in_duration = outlier_in_picked[
            #         outlier_in_picked.index.max() - maximum_outlier_duration:\
            #         outlier_in_picked.index.max()]
            #     if outlier_in_picked.shape[0]\
            #         and (outlier_in_duration==1).all()\
            #         and outlier_in_duration.size >= maximum_outlier_count:

            #         outlier_in_picked[outlier_in_duration.index] = 2
            #         outlier[outlier_in_duration.index] = 2

            # outlier_in_picked_index =\
            #     outlier_in_picked[outlier_in_picked == 1].index
            # # if outlier_in_picked_index.shape[0]:
            # #     import pdb
            # #     pdb.set_trace()
            # picked_data = picked_data.drop(index=outlier_in_picked_index)
            # while picked_data.size < search_parameters['max_count']\
            #     and current_bandwith < search_parameters['bandwith']['end']:
            #     current_bandwith += search_parameters['bandwith']['step']
            #     #get data
            #     picked_data = time_series[
            #         (time_series.index >= i - current_bandwith)\
            #             & (time_series.index < i)]
            #     outlier_in_picked =\
            #         outlier[outlier.index.isin(picked_data.index)]
            #     outlier_in_picked_index =\
            #         outlier_in_picked[outlier_in_picked == 1].index
            #     picked_data = picked_data.drop(index=outlier_in_picked_index)
                
            # if picked_data.size <= 100:
            #     outlier[ss.index] = 0
            # else:
            #     value = picked_data.values.reshape((-1,1))
            #     value_mean = value.mean()
            #     value_std = value.std()
                
            #     #avg mean
            #     outlier_bool =\
            #         np.abs(ss.values - value_mean)\
            #         > (standard_deviation_mutiplier * value_std)
            #     #idw mean
            #     # coord = picked_data.index\
            #     #     .astype('int64')\
            #     #     .astype('float')\
            #     #     .values.reshape((-1,1))
            #     # coord_est = ss.index\
            #     #     .astype('int64')\
            #     #     .astype('float')\
            #     #     .values.reshape((-1,1))
            #     # res = idw_est_coord_value(
            #     #     coord, value, coord_est,
            #     #     **inverse_distance_weighted_parameters)
            #     # outlier_bool =\
            #     #     np.abs(ss.values - res.flatten())\
            #     #     > (standard_deviation_mutiplier * value_std)

            #     outlier.loc[ss[outlier_bool].index] = 1
        else:
            pass
        if (idx+1) % print_step ==0:
            print '{f} -- {a}/{b}, data count: {c}, hour: {d}'.format(
                    f=time_series.name, a=idx+1, b=gp_len,
                    c=picked_data.shape[0], d=current_bandwith)
    print '{f} -- {a}/{b}, data count: {c}, hour: {d}'.format(
        f=time_series.name, a=idx+1, b=gp_len,
        c=picked_data.shape[0], d=current_bandwith)
    return outlier

def spatial_data_outlier_detection(
    spatial_data, estimation_boundary=['xmin', 'ymin', 'xmax', 'ymax'],
    standard_deviation_mutiplier=3,
    search_parameters={
        'bandwith': 1000,
        'max_count': 100
        },
    inverse_distance_weighted_parameters={
        'power': 1.5
        },
    print_step=2000):

    if estimation_boundary:
        estimation_spatial_data = spatial_data.loc[
        (spatial_data['x'] >= estimation_boundary[0]) &
        (spatial_data['x'] <= estimation_boundary[2]) &
        (spatial_data['y'] >= estimation_boundary[1]) &
        (spatial_data['y'] <= estimation_boundary[3])].copy()
    else:
        estimation_spatial_data = spatial_data.copy()
    estimation_spatial_data['outlier'] = False

    spatial_data_coord = spatial_data[['x', 'y']].values
    estimation_spatial_data_coord =\
        estimation_spatial_data[['x' ,'y']].values

    est_dict = neighbours_index_kd(
        estimation_spatial_data_coord,
        spatial_data_coord,
        nmax=search_parameters['max_count'],
        dmax=search_parameters['bandwith'])

    c_count = 0
    old_c_count = 0
    has_show_no_neighbor = False
    for c_idx, est_idx in est_dict.iteritems():
        # if 5852 in est_idx or 5936 in est_idx:
        #     import pdb
        #     pdb.set_trace()
        est_idx = np.array(est_idx, dtype=int)
        c_idx = np.array(c_idx, dtype=int)
        picked_c = spatial_data_coord[c_idx, :]
        picked_z = spatial_data[['z']].values[c_idx, :]
        
        if picked_c.shape[0] and np.unique(picked_c, axis=0).shape[0] == 1:
            if not has_show_no_neighbor:
                print 'some estimated point has no neighbors... default to False.'
                has_show_no_neighbor = True
            # estimation_spatial_data.loc[est_idx, 'z_est'] = np.nan
            # import pdb
            # pdb.set_trace()
            est_count = est_idx.shape[0]
            c_count+=est_count
            if c_count - old_c_count >= print_step:
                print c_count ,'/', estimation_spatial_data.shape[0]
                old_c_count = c_count
            continue
        else:
            value_mean = picked_z.mean()
            value_std = picked_z.std()

            picked_z_est = estimation_spatial_data[['z']].values[est_idx, :]
            # outlier_bool =\
            #     np.abs(picked_z_est.flatten() - value_mean)\
            #     > (standard_deviation_mutiplier * value_std)

            # idw mean
            picked_c_est = estimation_spatial_data_coord[est_idx, :]
            res = idw_est_coord_value(
                picked_c, picked_z, picked_c_est,
                **inverse_distance_weighted_parameters)
            if pd.isnull(res).any():
                print 'est has nan result... stop.'
                import pdb
                pdb.set_trace()
            outlier_bool =\
                np.abs(picked_z_est.flatten() - res.flatten())\
                > (standard_deviation_mutiplier * value_std)
            estimation_spatial_data.iloc[
                est_idx[outlier_bool],
                estimation_spatial_data.columns.get_loc('outlier')] = True
            est_count = est_idx.shape[0]
            c_count+=est_count
            if c_count - old_c_count >= print_step:
                print c_count ,'/', estimation_spatial_data.shape[0]
                old_c_count = c_count

    print c_count ,'/', estimation_spatial_data_coord.shape[0]
    return estimation_spatial_data
    
def space_time_outlier_detection(
    space_time_dataframe,
    id_column=None,
    spatial_column=None,
    temporal_column=None,
    value_column=None,
    space_group_frequency=None,
    space_standard_deviation_mutiplier=3,
    space_search_parameters={
        'bandwith': 10000,
        'max_count': 100},
    space_print_step=2000,
    time_group_frequency=None,
    time_standard_deviation_mutiplier=3,
    time_search_parameters={
        'bandwith':
            {'start': '2 hours',
             'step': '2 hours',
             'end': '2 hours'},
        'max_count': 100},
    time_print_step=2000,
    time_maximum_outlier_duration='1 hours',
    time_maximum_outlier_count=12,
    inverse_distance_weighted_parameters=None):

    def __time_series_outlier_detection(gpdf):
        print gpdf.name, '(', __time_series_outlier_detection.i+1,\
            '/', __time_series_outlier_detection.len, ')'
        ts = pd.Series(
            data=gpdf[value_column].values,
            index=gpdf[temporal_column],
            name=value_column)
        gg =\
        time_series_outlier_detection(
            time_series = ts, estimation_datetime_range = None,#['2016-12-30', '2017-3-1'],
            group_frequency=time_group_frequency,
            standard_deviation_mutiplier=3,
            maximum_outlier_duration=time_maximum_outlier_duration,
            maximum_outlier_count=time_maximum_outlier_count,
            search_parameters=time_search_parameters,
            inverse_distance_weighted_parameters=\
                inverse_distance_weighted_parameters,
            print_step=time_print_step)
        gpdf['temporal_outlier'] = gg.values
        __time_series_outlier_detection.i += 1
        return gpdf

    def __spatial_data_outlier_detection(gpdf):
        print gpdf.name, '(', __spatial_data_outlier_detection.i,\
            '/', __spatial_data_outlier_detection.len, ')'

        # {{for debug}}
        # if gpdf.name == pd.Timestamp('2016-10-20 07:10:00'):
        #     import ipdb
        #     ipdb.set_trace()
        # else:
        #     return gpdf
        gpdf = gpdf.rename(
            columns={
                spatial_column[0]: 'x',
                spatial_column[1]: 'y',
                value_column: 'z'})
        if gpdf.shape[0] == 0: #empty
            print 'no data in this group... skipped.'
            gpdf = pd.concat(
                [gpdf, pd.DataFrame(columns=['spatial_outlier'])]
                )
        else:
            gg =\
            spatial_data_outlier_detection(
                spatial_data = gpdf, estimation_boundary=None, #['xmin', 'ymin', 'xmax', 'ymax'],
                standard_deviation_mutiplier=3,
                search_parameters=space_search_parameters,
                inverse_distance_weighted_parameters=\
                    inverse_distance_weighted_parameters,
                print_step=space_print_step)
            gpdf['spatial_outlier'] = gg['outlier']
        __spatial_data_outlier_detection.i += 1
        return gpdf.rename(
            columns={
                'x': spatial_column[0],
                'y': spatial_column[1],
                'z': value_column})

    stdf = space_time_dataframe

    print 'start spatial data outlier...'
    #make space data outlier
    sdgp = stdf.groupby(pd.Grouper(
        key=temporal_column, freq=space_group_frequency))
    __spatial_data_outlier_detection.len = len(sdgp)
    __spatial_data_outlier_detection.i = 0
    stdf = sdgp.apply(__spatial_data_outlier_detection)

    print 'start time series outlier...'
    #make time series outlier
    tsgp = stdf.groupby(id_column)
    __time_series_outlier_detection.len = len(tsgp)
    __time_series_outlier_detection.i = 0
    stdf = tsgp.apply(__time_series_outlier_detection)

    stdf = stdf\
        .reset_index(drop=True)\
        .rename(
            columns={
                'x': spatial_column[0],
                'y': spatial_column[1],
                'z': value_column})
    stdf['outlier'] = stdf['temporal_outlier'] & stdf['spatial_outlier']
    #NOTE!!
    '''
    and tests whether both expressions are logically True
    while & (when used with True/False values) tests if both are True.
    '''


    # # set temporal event to False
    # stdf.loc[
    #     stdf['temporal_outlier']==2,
    #     "outlier"
    #     ] = False
    return stdf
    

    
    



# -*- coding: utf-8 -*-
from six.moves import range
import numpy as np

from ..general.neighbours import neighbours
from ..general.coord2K import coord2K
from ..stest.designmatrix import designmatrix

def kriging(ck,ch,zh,model,param,nhmax,dmax,order=None,options=None):
  ''' 
% kriging                   - prediction using kriging methods 
%
% Standard linear kriging algorithm for processing hard data
% that can reasonably be assumed as Gaussian distributed. The
% function is intendend to be as general as possible, covering
% various situations, like non-stationarity of the mean,
% multivariate cases, nested models, space-time estimations,
% etc. Depending on the case, specific format are needed for
% the input variables. This function is a special case of the more
% general BMEprobaMoments.mfunction, which processes both hard and 
% soft data.
%
% SYNTAX :
%
% [zk,vk]=kriging(ck,ch,zh,model,param,nhmax,dmax,order,options);
%
% INPUT :
%
% ck        nk by d   matrix of coordinates for the estimation locations.
%                     A line corresponds to the vector of coordinates at
%                     an estimation location, so the number of columns
%                     corresponds to the dimension of the space. There is
%                     no restriction on the dimension of the space.
% ch        nh by d   matrix of coordinates for the hard data locations,
%                     with the same convention as for ck.
% zh        nh by 1   vector of values for the hard data at the coordinates
%                     specified in ch.
  covmodel  list      list of m nested covariance models in each of 
                      which the spatial and temporal components are put 
                      in a list as [covmodelS,covmodelT] 
  covparam0 list      list of intial values covariance parameters for 
                      m covmodels in each component the parameters are 
                      listed as [sill,[covparamS1,covparamS2,..],
                      [covparamT1,covparamT2,..]]
% nhmax     scalar    maximum number of hard data values that are considered
%                     for the estimation at the locations specified in ck.
% dmax      scalar    maximum distance between an estimation location and
%                     existing hard data locations. All hard data locations
%                     separated by a distance smaller than dmax from an
%                     estimation location will be included in the estimation
%                     process for that location, whereas other hard data
%                     locations are neglected.
% order     scalar    order of the polynomial mean along the spatial axes at
%                     the estimation locations. For the zero-mean case, NaN
%                     (Not-a-Number) is used. Note that order=NaN can only be
%                     used with covariance models and not with variogram models.
% options   scalar    optional parameter that can be used if the default value
%                     is not satisfactory (otherwise it can simply be omitted
%                     from the input list of variables). options(1) is taking
%                     the value 1 or 0 if the user wants or does not want to
%                     display the order number of the location which is
%                     currently processed, respectively.
%
% OUTPUT :
%
% zk        nk by 1   vector of estimated values at the estimation locations. A
%                     value coded as NaN means that no estimation has been performed
%                     at that location due to the lack of available data. 
% vk        nk by 1   vector of estimation (kriging) variances at the estimation
%                     locations. As for zk, a value coded as NaN means that no
%                     estimation has been performed at the corresponding location.
%       
    
    
    input
    ck: nk by d float
    ch: nh by d float
    zh: nh by 1 float
    model: mn by 1 string
    param: mn by 3 float
    nhmax: int
    dmax: 1 by 3 float
    order: None or 0
    options: not no use here
    
    return
    zk: nk by 1 float
    vk: nk by 1 float
  '''
    
    #initial return variable
  nk = ck.shape[0]
  zk = np.zeros( ( nk, 1 ) )
  vk = np.zeros( ( nk, 1 ) )
  zk[:] = np.nan
  vk[:] = np.nan
  
  if order is None:
    order = np.nan
    
  for i in range( ck.shape[0] ):
    ci = ck[i:i+1]
    ci_nebr,zi_nebr,di_nebr,ni_nebr,idxi_nebr=neighbours(ci,ch,zh,nhmax,dmax)
    if ni_nebr > 0:
      K, dummyKK = coord2K( ci_nebr, ci_nebr, model, param )
      k, dummykk = coord2K( ci_nebr, ci, model, param ) #kk is a list contain ki
      k0, dummykk0 = coord2K(ci, ci, model, param)
      X=designmatrix(ci_nebr,order)[0]  
      x=designmatrix(ci,order)[0]
#            unit = np.ones(k.shape) # n by 1, X in matlab
#            unit_t_add = np.append( unit.T, [[0]], axis = 1 ) # 1 by n+1, x in matlab
                 
            #change shape for kriging
#            Kadd = np.hstack( ( K, unit ) )
#            Kadd = np.vstack( ( Kadd, unit_t_add ) )
      Kadd = K
      Xm,Xn = X.shape
      Kadd = np.vstack([np.hstack([K,X]),
                           np.hstack([X.T,np.zeros((Xn,Xn))])])
#            kadd = np.append( k, [[0]], axis = 0 )
      kadd = np.vstack([k,x.T])
      weight = np.linalg.solve(Kadd,kadd)[0:Xm]
      weight_t = weight.T
            #compute zk, vk
      zk[i] = weight_t.dot( zi_nebr )
      vk[i] = (k0-2*weight_t.dot(k)+weight_t.dot(K).dot(weight))[0]
            
    else:
      pass #already give NaN
  
  return zk, vk

def krigconstr(c0,c,order=None):
  '''
  order is None or 0 for now
  '''
  

  return X,x
    
def krigingFac( ck, ch, zh, model, param, nhmax, dmax, order = 0, options = None):

    ''' 
    input
    ck: nk by d float
    ch: nh by d float
    zh: nh by 1 float
    model: mn by 1 string ( ex. "exponential/gaussian" )
    param: mn by 3 float (ex. [ c, s, t ] )
    nhmax: int
    dmax: 1 by 3 float
    order: always 0, equals NaN in matlab
    options: not no use here, 0 or 1 in matlab for display echo 
    
    return
    zk: mn+1 by 1 float
    vk: mn+1 by 1 float
    '''
    
    #initial return variable
    nk = ck.shape[0]
    mn = len(model) #model.shape[0]
    zk = np.zeros( ( nk, mn+1 ) )
    vk = np.zeros( ( nk, mn+1 ) )
    zk[:] = np.nan
    vk[:] = np.nan
    
    for i in range( ck.shape[0] ):
        ci = ck[i:i+1]
        ci_nebr, zi_nebr, di_nebr, ni_nebr, idxi_nebr = neighbours(ci, ch, zh, nhmax, dmax )
        if ni_nebr > 0:
            K, dummyKK = coord2K( ci_nebr, ci_nebr, model, param )
            dummyk, kk = coord2K( ci_nebr, ci, model, param ) #kk is a list contain ki
            dummyk0, kk0 = coord2K(ci, ci, model, param)
            unit = np.ones(kk[0].shape) # n by 1
            unit_t_add = np.append( unit.T, [[0]], axis = 1 ) # 1 by n+1
            for idx_k, (ki,k0i) in enumerate( zip( kk, kk0 ) ):          
                #change shape for kriging
                Kadd = np.hstack( ( K, unit ) )
                Kadd = np.vstack( ( Kadd, unit_t_add ) )
                kkadd = np.append( ki, [[0]], axis = 0 )
                weight = np.dot( np.linalg.inv( Kadd ), kkadd )[:-1,:]
                weight_t = weight.T
                #compute zk, vk
                zk[i:i+1,idx_k:idx_k+1] = weight_t.dot( zi_nebr )
                vk[i:i+1,idx_k:idx_k+1] = ( k0i - 2 * weight_t.dot( ki ) + weight_t.dot( K ).dot( weight ) )[0]
            #compute local mean trend
            kkadd = np.append( np.zeros( kk[0].shape ), [[1]], axis = 0 )
            weight = np.dot( np.linalg.inv( Kadd ), kkadd )[:-1,:]
            weight_t = weight.T
            zk[i:i+1,-1:] = weight_t.dot( zi_nebr )
            vk[i:i+1,-1:] = weight_t.dot( K ).dot( weight )[0]
        else:
            pass #already give NaN
    return zk, vk




# -*- coding: utf-8 -*-
"""
Created on Thu Jul  2 22:16:03 2015

@author: hdragon689
"""
from six.moves import range
import numpy as np
import matplotlib.pyplot as plt
from mpl_toolkits.mplot3d import Axes3D
from ..stats.stcovfit import covmodelest,anisocovmodelest
from ..general.coord2K import coord2dist
from ..stats.stcovfit import cal_cov_mod

def modelplot(C,r,covmodel=None,covparam=None):
  '''
  Plot the empirical and modeled covariances for both spatial and S/T cases
  
  Syntax: modelplot(C,rLag,tLag,covmodel,covparam)

  Input:
  C     ns by nt    2D array of empirical covariance with ns spatial lags 
                    and nt temporal lags
  r     list        list of two 1D arrays of spatial and temporal lags 
                    [rLags, tLags]
  covmodel  list    default is None that only plot the empirical covariance                             
  covparam  list        
  
  Remark: detais of covmodel and covparam can refer to stamps.general.coord2K
  '''
  if len(r)==2 and type(r) is list: # space-time case
    rLag=r[0]
    tLag=r[1] 
    tLagM,rLagM=np.meshgrid(tLag,rLag)  
    if (covmodel is not None) and (covparam is not None):    
      rLagI=np.linspace(rLag[0],rLag[-1],50)
      tLagI=np.linspace(tLag[0],tLag[-1],50)   
      tLagMI,rLagMI=np.meshgrid(tLagI,rLagI)      
      modelcov,covi=covmodelest(rLagMI,tLagMI,covmodel,covparam)
    else:
      modelcov=np.zeros(1)*np.nan
      # Create cov plot and cov fit functions
    plt.figure()
    ax1=plt.subplot2grid((2,2),(0,0))
    ax2=plt.subplot2grid((2,2),(1,0))
    ax3=plt.subplot2grid((2,2),(0,1),rowspan=2,projection='3d')
    ax1.plot(rLag,C[:,0],'bo')
    ax2.plot(tLag,C[0,:],'bo') 
    ax3.scatter(rLagM,tLagM,C, color='b')
    ax1.set_xlabel('Spatial distance')
    ax1.set_ylabel('Covariance')
    ax2.set_xlabel('Temporl lag')
    ax2.set_ylabel('Covariance')
    ax3.set_xlabel('r')
    ax3.set_ylabel(r'$\tau$')
    ax3.set_zlabel('C')
    ax1.set_title('Spatial Covariance')
    ax2.set_title('Temporal Covariance')
    ax3.set_title('S/T Covariance')
    ax3.set_xlim([rLag[0],rLag[-1]])
    ax3.set_ylim([tLag[0],tLag[-1]])
    Cint=(C[~np.isnan(C)].max()-C[~np.isnan(C)].min())*0.02
    ax3.set_zlim([C[~np.isnan(C)].min()-Cint,C[~np.isnan(C)].max()+Cint])
    if not np.isnan(modelcov).all():
      ax1.plot(rLagI,modelcov[:,0],'r-')
      ax2.plot(tLagI,modelcov[0,:],'r-')
      ax3.plot_wireframe(rLagMI, tLagMI, modelcov,color='r', rstride=3, cstride=3)
      maxz=np.max([modelcov.max(),C[~np.isnan(C)].max()])
      minz=np.max([modelcov.min(),C[~np.isnan(C)].min()])
      ax3.set_zlim([minz,maxz])
      ax3.view_init(15,45); plt.draw()      
  else: 
    if len(r)==1 and type(r) is list:
      rLag=r[0].flat[:]
    elif type(r) is np.ndarray:
      rLag=r
    if (covmodel is not None) and (covparam is not None):    
      rLagI=np.linspace(rLag[0],rLag[-1],50)
      tLagI=np.array([0])
      tLagMI,rLagMI=np.meshgrid(tLagI,rLagI)  
      modelcov,covi=covmodelest(rLagMI,tLagMI,covmodel,covparam)
    plt.figure()
    plt.hold(True)
    plt.plot(rLag,C.flat[:],'o',label='Empirical covariance')
    if (covmodel is not None) and (covparam is not None):
      plt.plot(rLagI,modelcov.flat[:],'r-',label='Covariance model')
    plt.xlabel('Distance')
    plt.ylabel('Covariance')
    plt.xlim([rLag[0],rLag[-1]])
    Cint=(C[~np.isnan(C)].max()-C[~np.isnan(C)].min())*0.02
    plt.ylim([C[~np.isnan(C)].min()-Cint,C[~np.isnan(C)].max()+Cint])
    if (covmodel is not None) and (covparam is not None):
      plt.legend(loc='upper right')
  plt.show()    


def anisomodelplot(C,r,covmodel=None,covparam=None,theta=None,ratio=None):  
  '''
  Plot the empirical and modeled covariances
  
  Syntax: anisomodelplot(C,rLag,tLag,covmodel,covparam)

  Input:
  C     ns by nt    2D array of empirical covariance with ns spatial lags 
                    and nt temporal lags
  r     list        list of three 1D arrays of spatial and temporal lags 
                    [rLags, tLags, angles]. In pure spatial case, [rLags, angles] 
                    is used
  covmodel  list    default is None that only plot the empirical covariance                             
  covparam  list    
  theta     scalar  the theta for the covariance model (in radian)
  ratio     scalar  the ratio between secondary and principle axis of the 
                    anisotropic ellipse. The value is in the range between 0 and 1
  
  Remark: 
  detais of covmodel and covparam can refer to stamps.general.coord2K
  details of anisotropic parameters can refer to stamps.stats.stcovfit.anisocovmodelest
  
  '''  
  nang=len(C)
  if len(r)==3:
    rLag=r[0]
    tLag=r[1]
    aLag=r[2]  
  elif len(r)==2:
    rLag=r[0]
    aLag=r[1]
    tLag=None    
    
  modelcov=[None]*nang

  if covmodel is not None:
    # some error check may be required
    for i in range(nang):
      ang=np.array([aLag[i]])
      if tLag is None:
        tLag=np.array([0])
      tLagI,rLagI=np.meshgrid(tLag,rLag)
      covang,_= anisocovmodelest(covmodel,covparam,theta,ratio,ang,rLagI,tLagI)
      modelcov[i]=covang[0]

  if tLag is None or tLag.size == 1: # pure spatial case
    ax=[None]*nang
    cols=2
    rows=np.int(np.ceil(nang/cols))
    plt.figure()
    for i in range(nang):
      if i<rows:
        ax[i]=plt.subplot2grid((rows,cols),(i,0))
      else:
        ax[i]=plt.subplot2grid((rows,cols),(i-rows,1))
      ax[i].plot(rLag,C[i][:,0],'bo')
      ax[i].set_xlabel('Spatial distance')
      ax[i].set_ylabel('Covariance')
      ax[i].set_xlim([rLag[0],rLag[-1]])
      ax[i].text(0.95, 0.95, '$%5.2f^o$' % (aLag[i]/np.pi*180.),
            verticalalignment='top', horizontalalignment='right',
            transform=ax[i].transAxes,color='black', fontsize=10)      
      if modelcov[0] is not None:
        ax[i].plot(rLag,modelcov[i][:,0],'r-')
      if i==0 or i-rows==0:
        ax[i].set_title('Spatial Covariance')     
  else:
    ax=[None]*nang*2
    cols=2
    rows=nang
    plt.figure()
    for i in range(nang):
      ax[i]=plt.subplot2grid((rows,cols),(i,0))
      ax[i].plot(rLag,C[i][:,0],'bo')
      ax[i+rows]=plt.subplot2grid((rows,cols),(i,1))
      ax[i+rows].plot(tLag,C[i][0,:],'bo')
      ax[i].text(0.95, 0.95, '$%5.2f^o$' % (aLag[i]/np.pi*180.),
        verticalalignment='top', horizontalalignment='right',
        transform=ax[i].transAxes,color='black', fontsize=10)
      ax[i+rows].text(0.95, 0.95, '$%5.2f^o$' % (aLag[i]/np.pi*180.),
        verticalalignment='top', horizontalalignment='right',
        transform=ax[i+rows].transAxes,color='black', fontsize=10)  
      if i==0:
        ax[i].set_title('Spatial Covariance' % (aLag[i]/np.pi*180.))  
        ax[i+rows].set_title('Temporal Covariance' % (aLag[i]/np.pi*180.))               
      if i==nang-1:
        ax[i].set_xlabel('Spatial distance')
        ax[i].set_ylabel('Covariance')
        ax[i].set_xlim([rLag[0],rLag[-1]])
        ax[i+rows].set_xlabel('Temporal lag')
        ax[i+rows].set_ylabel('Covariance')
        ax[i+rows].set_xlim([tLag[0],tLag[-1]])
      else:
        ax[i].axes.get_xaxis().set_ticks([])
        ax[i+rows].axes.get_xaxis().set_ticks([])

      if modelcov[0] is not None:
        ax[i].plot(rLag,modelcov[i][:,0],'r-')    
        ax[i+rows].plot(tLag,modelcov[i][0,:],'r-') 


def mlemodelplot(ch,zh,covmodel=None,covparam=None):
  
  zh=zh.reshape(zh.size,1)
 
  cov=zh.dot(zh.T)
  covu=cov[np.triu_indices(zh.size)]
  dist_s=coord2dist(ch[:,:2],ch[:,:2])
  distu_s=dist_s[np.triu_indices(zh.size)]
  if ch.shape[1]>2:
    if type(ch[0,-1])==np.datetime64:
      origin=ch[0,-1]
      ch[:,-1]=np.double(np.asarray(ch[:,-1],dtype='datetime64')-origin)
      ch=ch.astype(np.double)
    dist_t=coord2dist(ch[:,2],ch[:,2])
    distu_t=dist_t[np.triu_indices(zh.size)]
    
  if (covmodel is not None) and (covparam is not None):
    rLagI=np.linspace(0,dist_s.max()*2./3,50)
    if ch.shape[1]>2:
      tLagI=np.linspace(0,dist_t.max()*2./3,50)   
    else:
      tLagI=np.array([0])
    tLagMI,rLagMI=np.meshgrid(tLagI,rLagI)      
    modelcov,covi=covmodelest(rLagMI,tLagMI,covmodel,covparam)  
    
  if ch.shape[1]>2:
    print ''
    # to be written
  else:
    plt.figure()
    plt.hold(True)
    plt.scatter(distu_s.flat[:],covu.flat[:])
    plt.plot(rLagI,modelcov,'r-')
    plt.xlim(0,rLagI.max())
    plt.ylim(covu.min(),covu.max())
    plt.xlabel('Distance')
    plt.ylabel('Covariance')


def covariance_model_plot(cov_model, s_range=None, t_range=None, show=True):
    s_range = max([i[2] for i in cov_model]) if not s_range else s_range
    t_range = max([i[4] for i in cov_model]) if not t_range else t_range

    cov_s_lags = np.linspace(0, s_range, 100)
    cov_t_lags = np.linspace(0, t_range, 100)
    cov_t_lags_grid, cov_s_lags_grid = np.meshgrid(cov_t_lags,cov_s_lags)


    cov_st_lags = np.array(
        zip(*map(
            lambda x:x.flatten(),
            np.meshgrid(cov_s_lags,cov_t_lags)
            ))
        )
    cov_z = cal_cov_mod(cov_st_lags, cov_model)
    cov_z_grid = cov_z.reshape((cov_s_lags.shape[0],cov_t_lags.shape[0]))

    if show:
        plt.figure(1)
        plt.subplot(211)
        plt.plot(cov_s_lags_grid[:,0], cov_z_grid[:,0], 'b--')
        plt.subplot(212)
        plt.plot(cov_t_lags_grid[0], cov_z_grid[0], 'r--')

        from mpl_toolkits.mplot3d import Axes3D
        fig = plt.figure(2)
        ax3d = Axes3D(fig)
        ax3d.plot_wireframe(cov_t_lags_grid, cov_s_lags_grid, cov_z_grid)

    
        plt.show()

    return cov_t_lags_grid, cov_s_lags_grid, cov_z_grid



# -*- coding:utf-8 -*-
import numpy as np
import multiprocessing as mp
CPU_COUNT = mp.cpu_count()

from six.moves import range
from scipy import sparse
from scipy.spatial import cKDTree as KDTree
from scipy.spatial.distance import cdist


def coord2dist_mp( c1, c2, workers=max(1, CPU_COUNT - 1) ):
    '''
    workers    int    number of workers
    '''

    if c1.shape[0] * c2.shape[0] <= 10 ** 7 or workers == 1:
        return coord2dist(c1, c2)
    else:
        print('used worker: {w}'.format(w = workers))
        d_c1 = np.int(np.ceil(c1.shape[0] / float(workers)))
        d_c2 = np.int(np.ceil(c2.shape[0] / float(workers)))
        c12 = [(c1[d_c1 * i:d_c1 * (i + 1)],
                c2[d_c2 * i:d_c2 * (i + 1)]) for i in range(workers)]
        res_ij = [(c12[i][0], c12[j][1]) for i in range(workers)
                  for j in range(workers)]
        pool = mp.Pool(processes=workers)
        d = pool.map(_warp_coord2dist, res_ij)

        d = [np.vstack(d[i::workers]) for i in range(workers)]
        d = np.hstack(d)
        return d
    return d

def _warp_coord2dist(args):
    return coord2dist(*args)

def coord2dist( c1, c2, norm=2 ):
    '''
    Calculate the distance between coordinates c1 and c2
    
    Syntax: result = coord2dist(c1, c2)

    Input 
        c1        [r1 x d]     np.array of coordinates
        c2        [r2 x d]     np.array of coordinates
    Output
        d         [r1 x r2]    np.array of distances
    '''

    try:
        d = cdist(c1, c2, 'minkowski', norm)  
    except:
        ones_c1 = np.ones((c1.shape[0], 1))
        ones_c2 = np.ones((c2.shape[0], 1))
        a = np.kron(c1, ones_c2)
        b = np.kron(ones_c1, c2)
        d = ((a - b) ** 2)
        d = result.sum(axis=1)
        d = np.sqrt(d)
        d = d.reshape((c1.shape[0], c2.shape[0]))
    return d



#!/usr/bin/env python2
# -*- coding: utf-8 -*-

"""
Created on Sat Dec 17 17:27:52 2016

@author: hdragon689

Random field generation by circular embedding method 
"""



'''
function [field1,field2,tx,ty]=stationary_Gaussian_process(m,n,rho)
% simulating stationary Gaussian field over an 'm' times 'n' grid
% INPUT:   
%          - 'm' and 'n' for evaluating the field over the m*n grid;
%             note that size of covariance matrix is m^2*n^2;
%          - scalar function rho(h), where 'h' is a two dimensional vector  
%            input and cov(X_t,Y_s)=rho(t-s) is the cov. function of a  
%            2-dimensional stationary Gaussian field; see reference below;
% OUTPUT:  
%          - two statistically independent fields 'field1' and 'field2'
%            over the m*n grid;
%          - vectors 'tx' and 'ty' so that the field is plotted via
%                     imagesc(tx,ty,field1)             
% Example:  
% rho=@(h)((1-h(1)^2/50^2-h(1)*h(2)/(15*50)-h(2)^2/15^2)...
%  *exp(-(h(1)^2/50^2+h(2)^2/15^2))); % define covariance function
%  stationary_Gaussian_process(512,384,rho); % plot when no output wanted   

%% Reference:
% Kroese, D. P., & Botev, Z. I. (2015). Spatial Process Simulation.
% In Stochastic Geometry, Spatial Statistics and Random Fields(pp. 369-404)
% Springer International Publishing, DOI: 10.1007/978-3-319-10064-7_12

tx=[0:n-1]; ty=[0:m-1]; % create grid for field
Rows=zeros(m,n); Cols=Rows;
for i=1:n % sample covariance function at grid points;
    for j=1:m
        Rows(j,i)=rho([tx(i)-tx(1),ty(j)-ty(1)]); % rows of blocks of cov matrix
        Cols(j,i)=rho([tx(1)-tx(i),ty(j)-ty(1)]); % columns of blocks of cov matrix
    end
end
% create the first row of the block circulant matrix with circular blocks
% and store it as a matrix suitable for fft2;
BlkCirc_row=[Rows, Cols(:,end:-1:2);
    Cols(end:-1:2,:), Rows(end:-1:2,end:-1:2)];
% compute eigen-values
lam=real(fft2(BlkCirc_row))/(2*m-1)/(2*n-1);
if abs(min(lam(lam(:)<0)))>10^-15
    error('Could not find positive definite embedding!')
else
    lam(lam(:)<0)=0; lam=sqrt(lam);
end
% generate field with covariance given by block circular matrix
F=fft2(lam.*complex(randn(2*m-1,2*n-1),randn(2*m-1,2*n-1)));
F=F(1:m,1:n); % extract subblock with desired covariance
field1=real(F); field2=imag(F); % two independent fields with desired covariance
if nargout==0
    imagesc(tx,ty,field1), colormap bone
end
'''



# -*- coding: utf-8 -*-
import os
import sys
import numpy
try:
    import nlopt
except ImportError, e:
    print ('Cannot import modeul "nlopt", try to install "nlopt" first.')
    raise e


def bobyqa(obj_func, init_guess, args, low_bnd, up_bnd,
    stop_val=None, maxeval=3000,
    ftol_rel=None, ftol_abs=None,
    xtol_rel=None, xtol_abs=None):
    '''
    Optimization by bobyqa method
    
    Function:
        bobyqa(obj_func, init_guess, args, low_bnd, up_bnd, maxeval)
    
    Input: 
    obj_func      function
        objective function to be minimized
    init_guess    array
        a 1D numpy array for the parameters to be estimated
    args          array
        a 1D numpy array for the function's arguments
    low_bnd       array
        optional. a 1D numpy array for parameter's lower bound
    up_bnd        array
        optional. a 1D numpy array for parameter's upper bound
    maxeval       int
        optional. a integer for maximum evaluation
    '''

    opt = nlopt.opt(nlopt.LN_BOBYQA, len(init_guess))
    opt.set_min_objective(
        lambda x, grad = numpy.array([]): obj_func(x, *args)
        )
    if not (low_bnd is None):
        opt.set_lower_bounds(low_bnd)
    if not (up_bnd is None):  
        opt.set_upper_bounds(up_bnd)
    if stop_val:
        opt.set_stopval(stop_val)
    opt.set_maxeval(maxeval)
    if ftol_rel:
        opt.set_ftol_rel(ftol_rel)
    if ftol_abs:
        opt.set_ftol_abs(ftol_abs)
    if xtol_rel:
        opt.set_xtol_rel(xtol_rel)
    if xtol_abs:
        opt.set_xtol_abs(xtol_abs)
    try:
        result = opt.optimize(init_guess)
    except nlopt.RoundoffLimited:
        print 'STAMPS Warning: roundoff limited found. Set relative value to 10**-8'
        opt.set_ftol_rel(10**-8)
        result = opt.optimize(init_guess)
    opt_val = opt.last_optimum_value()
    return result, opt_val



# -*- coding: utf-8 -*-
import copy

from six.moves import range
from six import print_ as print
from six import iteritems
import numpy as np
import scipy.stats
from scipy.spatial.distance import pdist
from scipy.spatial import cKDTree

from .softconverter import proba2stat
from .pystks_variable import get_standard_order, get_standard_soft_pdf_type
from .BMEoptions import BMEoptions
from .softconverter import pdf2cdf

from ..general.coord2K import coord2K, coord2Ksplit, coord2Kcombine
from ..stats.mepdf import maxentpdf_gc, maxentcondpdf_gc
from ..general.valstvgx import valstv2stg, valstg2stv
from ..general.neighbours import neighbours, neighbours_index_kd
from ..mvn.qmc import qmc

KHS_DICT = {'k': 0, 'h': 1, 's': 2}


def _bme_posterior_pdf(
    ck, ch=None, cs=None, zh=None, zs=None,
    covmodel=None, covparam=None, covmat=None,
    order=np.nan, options=None,
    general_knowledge='gaussian',
    #  specific_knowledge='unknown',  
    pdfk=None, pdfh=None, pdfs=None, hk_k=None, hk_h=None, hk_s=None,
    gui_args=None):

    def __get_zs_integration_limits(zs):
        '''get soft data integration limit''' 
        ranges = []
        for zsi in zs:
            pdftype = get_standard_soft_pdf_type(zsi[0])
            if pdftype in  [1, 2]: # nl, limi, probdens
                nl = zsi[1]
                limi = zsi[2]
                ranges.append((limi[0], limi[nl[0]-1]))
            elif pdftype in [10]:
                zm = zsi[1]
                zstd = np.sqrt(zsi[2])
                ranges.append((zm-3*zstd, zm+3*zstd))
        ranges = np.array(ranges)
        return ranges.copy()
    order = get_standard_order(order)
    nk = ck.shape[0]
    nh = ch.shape[0] if ch is not None else 0
    ns = cs.shape[0] if cs is not None else 0

    x_all_split = _get_x_all_split(nk, zh, zs)
    Xh = _get_x(x_all_split, 'h')
    mean_all_split = _get_mean_all_split(x_all_split, order)
    #get cov_all_split
    if covmat is None:
        cov_all_split = _get_cov_all_split(ck, ch, cs, covmodel, covparam)
        covmat = coord2Kcombine(cov_all_split)
    else:
        cov_all_split = np.vsplit(
            covmat, [nk, nk+nh, nk+nh+ns]
            )[:-1] #exclude final empty array
        cov_all_split = \
            [np.hsplit(c, [nk, nk+nh, nk+nh+ns][:-1])\
            for c in cov_all_split]

    #find cuplicated point
    if ns:
        dup_ck_cs_idx = np.array(
            np.all((ck[:, None, :] == cs[None, :, :]), axis=-1).nonzero()
            ).T
    else:
        dup_ck_cs_idx = np.array([[]])

    if general_knowledge == 'gaussian':
        
        def __get_fG_Xkh_each_ck(fG_Xkh_):
            if Xh is None:
                def fG_Xkh_each_ck(xk):
                    xk_origin_shape = xk.shape
                    xk = xk.flatten()
                    input_xk = xk.T
                    return fG_Xkh_(input_xk).reshape(xk_origin_shape)
            else:
                def fG_Xkh_each_ck(xk):
                    xk_origin_shape = xk.shape
                    xk = xk.flatten()
                    input_xk = np.vstack(
                        (xk, np.tile(Xh, xk.size))
                        ).T
                    return fG_Xkh_(input_xk).reshape(xk_origin_shape)
            return fG_Xkh_each_ck     
        
        #fG_Xh: const
        if nh == 0:
            fG_Xh = 1.
        else:
            fG_Xh = _get_multivariate_normal_pdf(
                x_all_split, mean_all_split, cov_all_split, 'h')(
                    _get_x(x_all_split, 'h').T)

        def __get_fG_Xs_gvn_Xh_all_ck():
            fG_Xs_gvn_Xh = _get_multivariate_normal_pdf(
                x_all_split, mean_all_split, cov_all_split, 's_h')
            def fG_Xs_gvn_Xh_all_ck(xs):
                xs_origin_shape = xs.shape
                xs = xs.reshape(-1, xs_origin_shape[-1])
                output = fG_Xs_gvn_Xh(xs)
                return output.reshape(xs_origin_shape[:-1]+(1,))
            return fG_Xs_gvn_Xh_all_ck
        fG_Xs_gvn_Xh = __get_fG_Xs_gvn_Xh_all_ck()

        zs_limits = __get_zs_integration_limits(zs)
        if options['integration method'] == 'qmc':
            fS_Xs = _get_fs(zs)
            def __qmc_int_fG_Xs_gvn_Xh__fS_Xs(x_array):
                return fG_Xs_gvn_Xh(x_array) * fS_Xs(x_array)
            xmin = zs_limits[:,0].copy()
            xmax = zs_limits[:,1].copy()

        elif options['integration method'] == 'qmc_F':
            fS_Xs = _get_fs(zs)
            Fsinv = _get_Fsinv(zs)

            def __qmc_int_fG_Xs_gvn_Xh__fS_Xs(Fx_array):
                return fG_Xs_gvn_Xh(Fsinv(Fx_array))
            xmin = np.zeros(zs_limits[:,0].shape)
            xmax = np.ones(zs_limits[:,1].shape)

        elif options['integration method'] == 'qmc_T':
            from scipy.special import erfinv
            m = _get_mean_a_given_b(x_all_split, mean_all_split,
                                cov_all_split, 's', 'h')
            v = _get_sigma_a_given_b(cov_all_split, 's', 'h')
            u,s,vh = np.linalg.svd(v)
            A = vh.T*np.sqrt(s)

            fS_Xs = _get_fs(zs)
            def __qmc_int_fG_Xs_gvn_Xh__fS_Xs(Fx_array):
                Fx_array[Fx_array==0.] = 10**-5
                Fx_array[Fx_array==1.] = 1 - 10**-5

                x_array = np.sqrt(2) * erfinv(2*Fx_array - 1)
                xs_array = (A.dot(x_array.T) + m).T
                return fS_Xs(xs_array)
            xmin = np.zeros(zs_limits[:,0].shape)
            xmax = np.ones(zs_limits[:,1].shape)

        int_fG_Xs_gvn_Xh__fS_Xs, e, info = qmc(
            __qmc_int_fG_Xs_gvn_Xh__fS_Xs,
            xmin, xmax, relerr=options[3,0], pow2min=10,
            showinfo=options['qmc_showinfo']
            )

        if int_fG_Xs_gvn_Xh__fS_Xs == 0: # NC is zero
            print('warning: normolized constant is equal to 0.')
            #int_fG_Xs_gvn_Xh__fS_Xs: const

        def __get_fSk_Xk_each_ck(zsk):
            def fSk_Xk_each_ck(xk):
                if zsk is None:
                    return np.ones(xk.shape)
                else:
                    xk_origin_shape = xk.shape
                    xk = xk.flatten()
                    pdf_type = get_standard_soft_pdf_type(zsk[0])
                    if pdf_type == 2:
                        nl = zsk[1][0]
                        limi = zsk[2]
                        probdens = zsk[3]
                        y_i = np.interp(
                            xk, limi[:nl], probdens[:nl],
                            left = 0., right = 0.)
                    elif pdf_type == 1:
                        nl = zs[1]
                        limi = zs[2]
                        probdens = zs[3]
                        idd = np.where(xk - limi[:nl] > 0)[0]
                        y_i = probdens[idd]
                    elif pdf_type == 10:
                        zm = zsk[1]
                        zstd = np.sqrt(zsk[2])
                        try:
                            y_i = scipy.stats.norm.pdf(
                                xk, loc=zm, scale=zstd)
                        except FloatingPointError:
                            import pdb
                            pdb.set_trace()
                      
                    return y_i.reshape(xk_origin_shape)
            return fSk_Xk_each_ck

        def __get_fG_Xs_gvn_Xkh_each_ck(ck_i, cs, zs,
            x_all_split_each_ck, mean_all_split_each_ck, cov_all_split_each_ck):
            idx_result = np.where(np.all(ck_i == cs, axis=1))[0]
            if idx_result.size == 0:
                x_all_split_each_ck_dup = x_all_split_each_ck
                mean_all_split_each_ck_dup = mean_all_split_each_ck
                cov_all_split_each_ck_dup = cov_all_split_each_ck
            elif idx_result.size == 1:
                x_all_split_each_ck_dup =\
                    x_all_split_each_ck[:2] +\
                    [np.delete(x_all_split_each_ck[2], idx_result, axis=0)]
                mean_all_split_each_ck_dup =\
                    mean_all_split_each_ck[:2] +\
                    [np.delete(mean_all_split_each_ck[2], idx_result, axis=0)]
                cov_all_split_each_ck_dup = copy.deepcopy(cov_all_split_each_ck)
                cov_all_split_each_ck_dup[0][2] =\
                    np.delete(cov_all_split_each_ck[0][2], idx_result, axis=1)
                cov_all_split_each_ck_dup[1][2] =\
                    np.delete(cov_all_split_each_ck[1][2], idx_result, axis=1)
                cov_all_split_each_ck_dup[2][0] =\
                    np.delete(cov_all_split_each_ck[2][0], idx_result, axis=0)
                cov_all_split_each_ck_dup[2][1] =\
                    np.delete(cov_all_split_each_ck[2][1], idx_result, axis=0)
                cov_all_split_each_ck_dup[2][2] =\
                    np.delete(cov_all_split_each_ck[2][2], idx_result, axis=0)
                #be careful below
                cov_all_split_each_ck_dup[2][2] =\
                    np.delete(cov_all_split_each_ck_dup[2][2], idx_result, axis=1)
            elif idx_result.size > 1: #strange
                raise ValueError('ck match cs twice. (strange)')
            def __get_fG_Xs_gvn_Xkh_each_ck_eack_xk(xk):
                fG_Xs_gvn_Xkh_container = []
                xk_origin_shape = xk.shape
                xk = xk.flatten()
                if options['integration method'] == 'qmc_T':
                    for xk_i in xk:
                        x_all_split_each_ck_dup[0] = np.array([[xk_i]])
                        m = _get_mean_a_given_b(
                            x_all_split_each_ck_dup,
                            mean_all_split_each_ck_dup,
                            cov_all_split_each_ck_dup, 's', 'kh')
                        v = _get_sigma_a_given_b(cov_all_split_each_ck_dup, 's', 'kh')
                        u, s, vh = np.linalg.svd(v)
                        A = vh.T*np.sqrt(s)

                        fG_Xs_gvn_Xkh_ = lambda xs, A=A, m=m: (A, m)

                        def fG_Xs_gvn_Xkh_each_ck_eack_xk(xs, ff):
                            xs_origin_shape = xs.shape
                            xs = xs.reshape(-1, xs_origin_shape[-1])
                            A, m = ff(xs)
                            xs[xs==0.] = 10**-5
                            xs[xs==1.] = 1 - 10**-5
                            xs2 = np.sqrt(2) * erfinv(2*xs - 1)
                            xs_array = (A.dot(xs2.T) + m).T
                            return xs_array

                        fG_Xs_gvn_Xkh_container.append(
                            lambda xs, ff=fG_Xs_gvn_Xkh_: fG_Xs_gvn_Xkh_each_ck_eack_xk(xs, ff))
                else:
                    for xk_i in xk:
                        x_all_split_each_ck_dup[0] = np.array([[xk_i]])
                        fG_Xs_gvn_Xkh_ = _get_multivariate_normal_pdf(
                        x_all_split_each_ck_dup,
                        mean_all_split_each_ck_dup,
                        cov_all_split_each_ck_dup, 's_kh')
                        def fG_Xs_gvn_Xkh_each_ck_eack_xk(xs, ff):
                            xs_origin_shape = xs.shape
                            xs = xs.reshape(-1, xs_origin_shape[-1])
                            output = ff(xs)
                            return output.reshape(xs_origin_shape[:-1]+(1,))
                        fG_Xs_gvn_Xkh_container.append(
                            lambda xs, ff=fG_Xs_gvn_Xkh_: fG_Xs_gvn_Xkh_each_ck_eack_xk(xs, ff))
                return np.array(
                    fG_Xs_gvn_Xkh_container).reshape(xk_origin_shape)
            return __get_fG_Xs_gvn_Xkh_each_ck_eack_xk

        def __get_fS_Xs_dup(ck_i, cs, zs):
            idx_result = np.where(np.all(ck_i == cs, axis=1))[0]
            if idx_result.size == 0:
                if options['integration method'] == 'qmc_F':
                    return Fsinv
                else:
                    return fS_Xs
            elif idx_result.size == 1:
                if options['integration method'] == 'qmc_F':
                    return _get_Fsinv(
                        [zs_i for i, zs_i in enumerate(zs) if i != idx_result])
                else:
                    return _get_fs(
                        [zs_i for i, zs_i in enumerate(zs) if i != idx_result])
            elif idx_result.size > 1: #strange
                raise ValueError('ck match cs twice. (strange)')

        def __get_pdf_each_ck(i):
            ck_i = ck[i]
            idx_result = np.where(np.all(ck_i == cs, axis=1))[0]
            if idx_result.size == 0:
                zs_dup = zs
            elif idx_result.size == 1:
                zs_dup = [zs_i for ii, zs_i in enumerate(zs) if ii != idx_result]
            elif idx_result.size > 1: #strange
                raise ValueError('ck match cs twice. (strange)')
            def _fK_Xk(xk):
                xk_origin_shape = xk.shape
                xk = xk.flatten()
                zs_limits = __get_zs_integration_limits(zs_dup)
                if options['integration method'] == 'qmc':
                    def __qmc_int_fG_Xs_gvn_Xkh_dup__fS_Xs_dup(x_array):
                        G_Xs_gvn_Xkh_dup_i =\
                            np.hstack(
                                [fi(x_array) for fi in fG_Xs_gvn_Xkh_dup[i](xk)])
                        fS_Xs_dup_i = fS_Xs_dup[i](x_array)
                        return G_Xs_gvn_Xkh_dup_i * fS_Xs_dup_i
                    xmin = zs_limits[:,0].copy()
                    xmax = zs_limits[:,1].copy()

                elif options['integration method'] == 'qmc_F':
                    def __qmc_int_fG_Xs_gvn_Xkh_dup__fS_Xs_dup(Fx_array):
                        G_Xs_gvn_Xkh_dup_i =\
                            np.hstack(
                                [fi(fS_Xs_dup[i](Fx_array)) for fi in fG_Xs_gvn_Xkh_dup[i](xk)])
                        return G_Xs_gvn_Xkh_dup_i
                    xmin = np.zeros(zs_limits[:,0].shape)
                    xmax = np.ones(zs_limits[:,1].shape)

                elif options['integration method'] == 'qmc_T':
                    def __qmc_int_fG_Xs_gvn_Xkh_dup__fS_Xs_dup(Fx_array):
                        G_Xs_gvn_Xkh_dup_i =\
                            [fi(Fx_array) for fi in fG_Xs_gvn_Xkh_dup[i](xk)]
                        fS_Xs_dup_i =\
                            np.hstack(
                                [fS_Xs_dup[i](xs_array_i) for xs_array_i in G_Xs_gvn_Xkh_dup_i] )
                        return fS_Xs_dup_i
                    xmin = np.zeros(zs_limits[:,0].shape)
                    xmax = np.ones(zs_limits[:,1].shape)

                int_fG_Xs_gvn_Xkh_dup__fS_Xs_dup , e, info = qmc(
                    __qmc_int_fG_Xs_gvn_Xkh_dup__fS_Xs_dup,
                    xmin, xmax, relerr=options[3,0], pow2min=10,
                    showinfo=options['qmc_showinfo']
                    )
                int_fG_Xs_gvn_Xkh_dup__fS_Xs_dup =\
                    int_fG_Xs_gvn_Xkh_dup__fS_Xs_dup.reshape(xk_origin_shape)
                xk = xk.reshape(xk_origin_shape)
                if options['ck pdf debug']:
                    print('fG_Xkh[i](xk): {p}'.format(p=fG_Xkh[i](xk)))
                    print('fSk_Xk[i](xk): {p}'.format(p=fSk_Xk[i](xk)))
                    print('int_fG_Xs_gvn_Xkh_dup__fS_Xs_dup: {p}'.format(p=int_fG_Xs_gvn_Xkh_dup__fS_Xs_dup))
                    print('fG_Xh: {p}'.format(p=fG_Xh))
                    print('int_fG_Xs_gvn_Xh__fS_Xs: {p}'.format(p=int_fG_Xs_gvn_Xh__fS_Xs))
                    
                return (fG_Xkh[i](xk) * fSk_Xk[i](xk)
                    * int_fG_Xs_gvn_Xkh_dup__fS_Xs_dup
                    / fG_Xh / int_fG_Xs_gvn_Xh__fS_Xs)
            return _fK_Xk

        cov_hs_range = list(range(nk, nk+nh+ns))
        
        fG_Xkh = [] # a list contains each ck's fG_Xkh
        fSk_Xk = [] # a list contains each ck's fSk_Xk
        fS_Xs_dup = []
        fG_Xs_gvn_Xkh_dup = [] # a list contains each ck's fG_Xs_gvn_Xkh_dup
        fK_Xk = []
        for i in range(nk):
            #get x/mean/cov_all_split at each ck point
            x_all_split_each_ck =\
                [x_all_split[0][i:i+1,:]] + x_all_split[1:]
            mean_all_split_each_ck =\
                [mean_all_split[0][i:i+1,:]] + mean_all_split[1:]
            covmat_each_ck =\
                covmat[np.ix_(
                    [i]+cov_hs_range,
                    [i]+cov_hs_range
                    )]
            cov_all_split_each_ck = np.vsplit(
                covmat_each_ck, [1, 1+nh, 1+nh+ns]
                )[:-1] #exclude final empty array
            cov_all_split_each_ck = \
                [np.hsplit(c, [1, 1+nh, 1+nh+ns][:-1])\
                for c in cov_all_split_each_ck]

            # get fG_Xkh each ck part
            fG_Xkh_ = _get_multivariate_normal_pdf(
                x_all_split_each_ck,
                mean_all_split_each_ck,
                cov_all_split_each_ck, 'kh')
            fG_Xkh.append(
                __get_fG_Xkh_each_ck(fG_Xkh_))

            # get fSk_Xk
            idx_result = np.where(np.all(ck[i] == cs, axis=1))[0]
            if idx_result.size == 0:
                fSk_Xk.append(__get_fSk_Xk_each_ck(None))
            elif idx_result.size == 1:
                fSk_Xk.append(__get_fSk_Xk_each_ck(zs[idx_result[0]]))
            elif idx_result.size > 1: #strange
                raise ValueError('ck match cs twice. (strange)')

            # get fS_Xs_dup
            fS_Xs_dup.append(__get_fS_Xs_dup(ck[i], cs, zs))

            # get fG_Xs_gvn_Xkh
            fG_Xs_gvn_Xkh_dup.append(
                __get_fG_Xs_gvn_Xkh_each_ck(ck[i], cs, zs,
                    x_all_split_each_ck,
                    mean_all_split_each_ck,
                    cov_all_split_each_ck))

            fK_Xk.append(__get_pdf_each_ck(i))

        return np.array([fK_Xk]).reshape((-1, 1))
    else: #general knowledge is not gaussian
        pass

def _bme_posterior_moments(
    ck, ch=None, cs=None, zh=None, zs=None,
    covmodel=None, covparam=None, covmat=None,
    order=np.nan, options=None,
    general_knowledge='gaussian',
    #  specific_knowledge='unknown',
    pdfk=None, pdfh=None, pdfs=None, hk_k=None, hk_h=None, hk_s=None,
    gui_args=None, ck_cov_output=False):

    '''
        no neighbour considered, so spatial-temporal range
        should be transform first (no dmax support).

        covmat:
            covariance matrix, a numpy 2d array
            with shape (nk+nh+ns) by (nk+nh+ns)
        if covmat provieded, covmodel and covparam are simply skipped.
    '''

    if general_knowledge == 'gaussian':
        if zs:
            all_zs_type = np.array(
                map(get_standard_soft_pdf_type, [zsi[0] for zsi in zs])
                )
            if (all_zs_type==10).all(): # all soft type are gaussian
                # find cs in ck index
                dup_index = np.where((ck==cs[:,None]).all(-1))[1]
                if dup_index.size > 0: #has duplicated point
                    if not ck_cov_output:
                        mask = np.ones(len(dup_index), dtype=bool)
                        mask[dup_index] = False
                        mvs = np.empty((ck.shape[0],3))
                        mvs[:] = np.nan
                        ck_dup = ck[dup_index, :]
                        ck_no_dup = np.delete(ck, dup_index, axis=0)
                        mvs[dup_index, :] =\
                            _bme_proba_gaussian_dup(
                                ck_dup, ch, cs, zh, zs,
                                covmodel, covparam, covmat, order, options)
                        if ck_no_dup.size > 0:
                            mvs[mask, :] =\
                                _bme_proba_gaussian(
                                    ck_no_dup, ch, cs, zh, zs,
                                    covmodel, covparam, covmat, order, options)
                        return mvs
                    else:
                        raise ValueError('We can not do ck_cov_output with duplicated point.')
                else:
                    if not ck_cov_output:
                        mvs = _bme_proba_gaussian(
                            ck, ch, cs, zh, zs,
                            covmodel, covparam, covmat, order, options)
                        return mvs
                    else:
                        mvs, ckcov = _bme_proba_gaussian(
                            ck, ch, cs, zh, zs,
                            covmodel, covparam, covmat, order, options, ck_cov_output)
                        return mvs, ckcov
            else: # has non-gaussian
                nk = ck.shape[0]
                nh = ch.shape[0] if ch is not None else 0
                ns = cs.shape[0] if cs is not None else 0
                x_all_split = _get_x_all_split(nk, zh, zs)
                mean_all_split = _get_mean_all_split(x_all_split, order)
                if covmat is not None:
                    cov_all_split = np.vsplit(
                        covmat, [nk, nk+nh, nk+nh+ns]
                        )[:-1] #exclude final empty array
                    cov_all_split = \
                        [np.hsplit(c, [nk, nk+nh, nk+nh+ns][:-1])\
                        for c in cov_all_split]
                else:
                    cov_all_split = _get_cov_all_split(
                        ck, ch, cs, covmodel, covparam)

                fg_s_given_h = _get_multivariate_normal_pdf(
                    x_all_split, mean_all_split,
                    cov_all_split, 's_h')
                fs = _get_fs(zs)

                #split hard and soft of m_k_gvn_hs
                m_k = _get_mean(mean_all_split, 'k')
                Bm_sigma_inv_multi = (
                    _get_sigma(cov_all_split, 'k', 'hs').dot(
                        _get_sigma(cov_all_split, 'hs', 'hs', inv=True))
                    )
                m_hs = _get_mean(mean_all_split, 'hs')
                m_k_gvn_hs_part = m_k - Bm_sigma_inv_multi.dot(m_hs)
                x_hs = _get_x(x_all_split, 'hs') #put true Xs later

                sigma_k_given_hs = _get_sigma_a_given_b(
                    cov_all_split, 'k', 'hs')
                diag_sigma_k_given_hs =\
                    np.diag(sigma_k_given_hs)

                def func_moments(x_array):
                    nMon = 3
                    npts = x_array.shape[0]
                    res = np.empty(
                        (npts, nMon*nk + 1)
                        )
                    x_hs_npts = np.tile(x_hs, (1, npts))
                    x_hs_npts[nh:,:] = x_array.T # put true Xs part

                    m_k_gvn_hs_npts = (
                        m_k_gvn_hs_part + Bm_sigma_inv_multi.dot(x_hs_npts)
                        ).T

                    fg_fs = (
                        fg_s_given_h(x_array).reshape((npts,1)) * fs(x_array)
                        )
                    res[:, 0*nk:1*nk] = m_k_gvn_hs_npts * fg_fs
                    res[:, 1*nk:2*nk] = (res[:,:nk] * m_k_gvn_hs_npts)
                    res[:, 2*nk:3*nk] = (
                        3 * diag_sigma_k_given_hs * res[:,:nk]
                        - 2 * res[:,nk:2*nk] * m_k_gvn_hs_npts
                        )
                    res[:,-1:] = fg_fs
                    return res

                ranges = []
                for zsi in zs:
                    pdftype = get_standard_soft_pdf_type(zsi[0])
                    if pdftype in  [1, 2]: # nl, limi, probdens
                        nl = zsi[1]
                        limi = zsi[2]
                        ranges.append((limi[0], limi[nl[0]-1]))
                    elif pdftype in [10]:
                        zm = zsi[1]
                        zstd = np.sqrt(zsi[2])
                        ranges.append((zm-5*zstd, zm+5*zstd))
                ranges = np.array(ranges)
                xmin = ranges[:,0].copy()
                xmax = ranges[:,1].copy()

                Mon, e, info = qmc(
                    func_moments, xmin, xmax, relerr=options[3,0],
                    showinfo=options['qmc_showinfo'])

                Mon_NC = Mon[-1]
                Mon = Mon[:-1].reshape((-1, nk)).T # k by nmon
                Mon /= Mon_NC # for moments 1,2,3
                Mon[:,1:2] += (
                    diag_sigma_k_given_hs.reshape((-1, 1))
                    ) # for moments 2

                M1 = Mon[:, 0:1]
                M2 = Mon[:, 1:2]
                M3 = Mon[:, 2:3]

                mvs = np.empty(Mon.shape)
                mvs[:,0:1] = M1
                mvs[:,1:2] = M2 - M1**2
                mvs[:,2:3] = M3 - 3*M2*M1 + 2*M1**3

                if not ck_cov_output:
                    return mvs
                else:
                    zs_gau = []
                    for zsi in zs:
                        zs_gau_m, zs_gau_v = proba2stat(
                            zsi[0],
                            np.array([zsi[1]]),
                            np.array([zsi[2]]),
                            np.array([zsi[3]])
                            )
                        zs_gau.append((10, zs_gau_m, zs_gau_v))
                    mvs22, ckcov22 = _bme_proba_gaussian(
                        ck, ch, cs, zh, zs_gau,
                        covmodel, covparam, covmat, order, options, ck_cov_output)

                    #do something transform between cov and corr

                    return mvs, ckcov
        else:
            if not ck_cov_output:
                mvs = _bme_proba_gaussian(
                    ck, ch, cs, zh, zs, covmodel, covparam, covmat, order, options)
                return mvs
            else:
                mvs, ckcov = _bme_proba_gaussian(
                    ck, ch, cs, zh, zs,
                    covmodel, covparam, covmat, order, options, ck_cov_output)
                return mvs, ckcov
    else:
        raise ValueError("Now we can not consider non-gaussian GK." )

def _bme_proba_gaussian(
    ck, ch=None, cs=None, zh=None, zs=None,
    covmodel=None, covparam=None, covmat=None,
    order=np.nan, options=None, ck_cov_output=False):
    '''
    no neighbour consider, no data format transform.

    ch, cs, zh: np.2darray or None
    zs: new zs data or None, see softconverter.py for detail.
    ck_cov_output: if True, result will additionally return
        covariance between ck
    NOTE zs there should gaussian type e.g.
        zs = ((10, mean1, var1),...,(10, mean2, var2))
    '''

    nk, nh, ns = __get_khs_size(ck, ch, cs)
    x_all_split, mean_all_split, cov_all_split, covmat =\
        __get_all_split(
            ck, ch, cs, zh, zs, covmodel, covparam, covmat, order)
    
    has_remove_index, remove_index =\
        __get_covmat_remove_index(covmat)

    if has_remove_index: #need change data
        s_remove_index = remove_index-(nk+nh)
        if (s_remove_index >= 0).all():
            cs = np.delete(cs, s_remove_index, axis=0)
            zs = [zs_i for idx, zs_i in enumerate(zs) if idx not in s_remove_index]
            covmat = np.delete(covmat, nk+nh+s_remove_index, axis=0)
            covmat = np.delete(covmat, nk+nh+s_remove_index, axis=1)

            x_all_split, mean_all_split, cov_all_split, covmat =\
                __get_all_split(
                ck, ch, cs, zh, zs, covmodel, covparam, covmat, order)
        else: # need remove k or h, strange
            import pdb
            pdb.set_trace()
    nk, nh, ns = __get_khs_size(ck, ch, cs)

    if ns == 0 and nh == 0:
        mvs = np.empty((ck.shape[0],3))
        mvs[:] = np.nan
        return mvs
        #raise ValueError('hard and soft data can not both without input')

    if ns == 0: # only hard data
        mean_k_given_h = _get_mean_a_given_b(
            x_all_split, mean_all_split,
            cov_all_split, sub_a='k', sub_b='h')
        sigma_k_given_h = _get_sigma_a_given_b(
            cov_all_split, sub_a='k', sub_b='h')

        skewness = np.zeros(mean_k_given_h.shape)
        mvs = np.hstack(
            (mean_k_given_h, sigma_k_given_h.diagonal().reshape((-1,1)),
            skewness)
            )
        if ck_cov_output:
            return mvs, sigma_k_given_h
        else:
            return mvs
    else: # both hard and soft data (hard data can be empty)
        mean_k = _get_mean(mean_all_split, 'k')
        
        #check outlier data
        mean_s_given_h = _get_mean_a_given_b(
            x_all_split, mean_all_split,
            cov_all_split, sub_a='s', sub_b='h')
        sigma_s_given_h = _get_sigma_a_given_b(cov_all_split, 's', 'h')
        general_mean = mean_s_given_h.ravel()
        general_var = np.diag(sigma_s_given_h)
        del sigma_s_given_h
        soft_mean = np.array([zs_i[1] for zs_i in zs])
        soft_var = np.array([zs_i[2] for zs_i in zs])

        problem_bool = (np.abs(general_mean - soft_mean)\
            > 3*(np.sqrt(general_var) + np.sqrt(soft_var))
            )
        if problem_bool.any():
            problem_index = np.where(problem_bool)[0]
            print('warning: the soft data at index(es) {i} '\
                'are far from its general marginal pdf'.format(
                    i=str(problem_index)))
            #remove and re calculate
            problem_index
            cs = np.delete(cs, problem_index, axis=0)
            zs = [zs_i for idx, zs_i in enumerate(zs) if idx not in problem_index]
            covmat = np.delete(covmat, nk+nh+problem_index, axis=0)
            covmat = np.delete(covmat, nk+nh+problem_index, axis=1)

            x_all_split, mean_all_split, cov_all_split, covmat =\
                __get_all_split(
                ck, ch, cs, zh, zs, covmodel, covparam, covmat, order)
            nk, nh, ns = __get_khs_size(ck, ch, cs)

            mean_s_given_h = _get_mean_a_given_b(
                x_all_split, mean_all_split,
                cov_all_split, sub_a='s', sub_b='h')

        mean_hs = _get_mean(mean_all_split, 'hs')

        NC, useful_args = _get_int_fg_a_given_b_fs_s(
            x_all_split, mean_all_split,
            cov_all_split, zs,
            sub_multi = 's_h', sub_s='s'
            )

        NC = __check_normalized_constant(NC, options)
        if NC == 0:
            print('mvs set to NaN.')
            mvs = np.empty((nk,3))
            mvs = np.nan
            return mvs
        if np.isnan(NC) or np.isinf(NC):
            print('NC error... is NaN or Inf.')
            if 1 or options['debug']: #always stop
                import pdb
                pdb.set_trace()

        (sigma_t_prime, inv_sigma_s_given_h,
            inv_sigma_tilde_s, mean_tilde_s,
            alias_c, alias_fgfs1234, alias_mean_d, alias_sigma_d
            ) = useful_args
        hat_x_s = NC * sigma_t_prime.dot(
            inv_sigma_s_given_h.dot(mean_s_given_h)
            + inv_sigma_tilde_s.dot(mean_tilde_s)
            )
        if not nh:
            hat_x_hs = hat_x_s
        else:
            hat_x_h = _get_x(x_all_split, 'h') * NC
            hat_x_hs = np.vstack((hat_x_h, hat_x_s))
        sigma_k_hs = _get_sigma(
            cov_all_split, sub_a='k', sub_b='hs')
        inv_sigma_hs_hs = _get_sigma(
            cov_all_split, sub_a='hs', sub_b='hs', inv=True)
        cond_k_hs = sigma_k_hs.dot(inv_sigma_hs_hs)
        BME_mean_k_given_hs_a = cond_k_hs.dot(hat_x_hs)
        BME_mean_k_given_hs_b = mean_k - cond_k_hs.dot(mean_hs)
        BME_mean_k_given_hs =\
            BME_mean_k_given_hs_a/NC + BME_mean_k_given_hs_b

        sigma_k_given_hs = _get_sigma_a_given_b(
            cov_all_split, sub_a='k', sub_b='hs')
        mean_t = hat_x_hs/NC
        aa = np.zeros(inv_sigma_hs_hs.shape)
        aa[nh:nh+ns, nh:nh+ns] = sigma_t_prime*NC
        bb = (mean_t - mean_hs).dot((mean_t - mean_hs).T) * NC
        tt = cond_k_hs.dot(aa + bb).dot(cond_k_hs.T)
        
        if not ck_cov_output:
            sigma_k_given_hs_diag =\
                sigma_k_given_hs.diagonal().reshape((-1,1))
            tt_diag = tt.diagonal().reshape((-1,1))
            BME_var_k_given_hs = (
                sigma_k_given_hs_diag - BME_mean_k_given_hs**2 + mean_k**2
                - 2*mean_k * cond_k_hs.dot(mean_hs)
                + 2*mean_k * cond_k_hs.dot(hat_x_hs) / NC
                + tt_diag / NC
                )
        else:
            exm = cond_k_hs.dot(hat_x_hs).dot(mean_k.T)
            emm = cond_k_hs.dot(mean_hs).dot(mean_k.T)
            BME_var_k_given_hs_cov = (
                sigma_k_given_hs
                - BME_mean_k_given_hs.dot(BME_mean_k_given_hs.T)
                + mean_k.dot(mean_k.T)
                + 2*exm - 2*emm
                + tt / NC
                )
            BME_var_k_given_hs =\
                BME_var_k_given_hs_cov.diagonal().reshape((-1,1))
            
        skewness = np.zeros(BME_mean_k_given_hs.shape)
        mvs = np.hstack(
            (BME_mean_k_given_hs, BME_var_k_given_hs, skewness)
            )
        if ck_cov_output:
            return mvs, BME_var_k_given_hs_cov
        else:
            return mvs

def _bme_proba_gaussian_dup(
    ck, ch=None, cs=None, zh=None, zs=None,
    covmodel=None, covparam=None, covmat=None,
    order=np.nan, options=None, ck_cov_output=False):

    nk, nh, ns = __get_khs_size(ck, ch, cs)
    x_all_split, mean_all_split, cov_all_split, covmat =\
        __get_all_split(
            ck, ch, cs, zh, zs, covmodel, covparam, covmat, order)

    if ns == 0: #strange, should have ck duplicates with cs
        raise ValueError('no cs, strange.')
    else:
        NC, (sigma_d, inv_sigma_s_given_h,
            inv_sigma_tilde_s, mean_tilde_s,
            alias_c, alias_fgfs1234, alias_mean_d, alias_sigma_d
            ) =\
            _get_int_fg_a_given_b_fs_s(
                x_all_split, mean_all_split, cov_all_split, zs, 
                sub_multi = 's_h', sub_s='s'
                )

        # no used but give a warning   
        #NC = __check_normalized_constant(NC)
        if NC == 0:
            print('warning: NC found to be 0.')
            if options['debug']:
                import pdb
                pdb.set_trace()
        # find ck in cs index
        ck_dup_index, cs_dup_index = np.where((cs==ck[:,None]).all(-1))
        return np.hstack((
            alias_mean_d[cs_dup_index,:],
            np.diagonal(alias_sigma_d).reshape((-1,1))[cs_dup_index,:],
            np.zeros((nk, 1))
            ))

        # NC, (sigma_d, inv_sigma_s_given_h,
        #     inv_sigma_tilde_s, mean_tilde_s,
        #     alias_c, alias_fgfs1234, alias_mean_d, alias_sigma_d
        #     ) =\
        #     _get_int_fg_a_given_b_fs_s(
        #         x_all_split, mean_all_split, cov_all_split, zs, 
        #         sub_multi = 's_h', sub_s='s'
        #         )
        # NC = __check_normalized_constant(NC)
        
        # #get top part
        # #ns x 1
        # top_part2 = NC * alias_mean_d
        # top_part = alias_c * alias_fgfs1234 * alias_mean_d
        # import pdb
        # pdb.set_trace()
        # #get each nc_dup
        # dup_NC = np.ones(top_part.shape)
        # # find ck in cs index
        # ck_dup_index, cs_dup_index = np.where((cs==ck[:,None]).all(-1))
        # for ck_i, cs_i in zip(ck_dup_index, cs_dup_index):
        #     #get x/mean/cov_all_split at each ck by remove dup cs point
        #     x_all_split_each_ck =\
        #         x_all_split[:2]\
        #         + [np.delete(x_all_split[2], cs_i, axis=0)]
        #     mean_all_split_each_ck =\
        #         mean_all_split[:2]\
        #         + [np.delete(mean_all_split[2], cs_i, axis=0)]

        #     covmat_each_ck = np.delete(covmat, nk+nh+cs_i, axis=0)
        #     covmat_each_ck = np.delete(covmat_each_ck, nk+nh+cs_i, axis=1)
           
        #     cov_all_split_each_ck = np.vsplit(
        #         covmat_each_ck, [nk, nk+nh, nk+nh+ns-1]
        #         )[:-1] #exclude final empty array
        #     cov_all_split_each_ck = \
        #         [np.hsplit(c, [nk, nk+nh, nk+nh+ns-1][:-1])\
        #         for c in cov_all_split_each_ck]
        #     zs_each_ck = [z for idx_z, z in enumerate(zs) if idx_z != cs_i]

        #     dup_NC[cs_i, 0] = __check_normalized_constant(
        #         __get_normalized_constant(
        #             x_all_split_each_ck,
        #             mean_all_split_each_ck,
        #             cov_all_split_each_ck, zs_each_ck, 
        #             sub_multi = 's_h', sub_s='s'
        #             )
        #         )
        # return (top_part/dup_NC)[cs_dup_index,:]

def _get_x_all_split(nk, zh, zs):
    '''
        Create the "estimated" observed values 
    for the estimation and observations
        For now, zero is used for estimation points
    (which should be specified as NaN)
        mean values are used for soft data
    '''
    x_all_split = []
    xk = np.empty((nk, 1))  # will be replaced later
    x_all_split.append(xk)
    x_all_split.append(zh) # e.g. xh
    if zs:
        xs = np.empty((len(zs), 1))
        for i, zsi in enumerate(zs):
            if get_standard_soft_pdf_type(zsi[0]) == 10:  # gaussian/normal
                xs[i] = zsi[1] #z_mean
            else:
                xs[i], dummy_v = proba2stat(
                    zsi[0],
                    np.array([zsi[1]]),
                    np.array([zsi[2]]),
                    np.array([zsi[3]])
                    )
        x_all_split.append(xs)
    else:
        x_all_split.append(None)
    return x_all_split

def _get_mean_all_split(x_all_split, order):
    '''
    Obtain the trend estimations at the estimation and data locations based 
    upon the specified trend order
    '''
    if isinstance(order, np.ndarray):  # user defined general knowledge, (row x 1 2d array)
        start_i = 0
        mean_all_split = []
        for i in x_all_split:
            if i is not None:
                end_i = start_i + i.size
                mean_all_split.append(order[start_i:end_i, :])
                start_i = end_i
            else:
                mean_all_split.append(None)   
    elif order == 0:  # constant mean, exclude zk, average h and s
        xx=[x for x in x_all_split[1:] if x is not None]
        constant_mean_ = np.vstack(xx).mean()
        mean_all_split = []
        for i in x_all_split:
            if i is not None:
                mean_all_split.append(np.ones(i.shape)*constant_mean_)
            else:
                mean_all_split.append(None)
        #for means in mean_all_split:
        #  means[:] = constant_mean_
    elif np.isnan(order):  # zero mean: 
        mean_all_split = []
        for i in x_all_split:
            if i is not None:
                mean_all_split.append(np.zeros(i.shape))
            else:
                mean_all_split.append(None)
    return mean_all_split

def _get_cov_all_split(ck, ch, cs, covmodel, covparam):
  '''
  Obtain the covariance in the split ways.
  See coord2K
  '''
  return coord2Ksplit((ck, ch, cs), (ck, ch, cs),
                      covmodel, covparam)[0]

def _get_x(x_all_split, sub):
  '''
  Retrieve the estimated observed values given specified class, i.e., sub
  sub can be k, h, and s for estimation, hard, and soft data
  '''
  idx = [KHS_DICT[i] for i in sub]
  output=[x_all_split[i] for i in idx if x_all_split[i] is not None]
  if len(output)>0:
    return np.vstack(output)
  else:
    return None

def _get_mean(mean_all_split, sub):
  '''
  Retrieve the expected values given specified class, i.e., sub
  sub can be k, h, and s for estimation, hard, and soft data
  '''
  idx = [KHS_DICT[i] for i in sub]     
  output=[mean_all_split[i] for i in idx if mean_all_split[i] is not None]
  if len(output)>0:
    return np.vstack(output)
  else:
    return None

def _get_sigma(cov_all_split, sub_a, sub_b, inv=False):
  '''
  Retrieve the cross-covaiance between specified class, i.e., sub_a and sub_b
  sub_a and sub_b can be k, h, and s for estimation, hard, and soft data
  '''
  idx_a = [KHS_DICT[i] for i in sub_a]
  idx_b = [KHS_DICT[i] for i in sub_b]

  cov_a_b = []
  for i in idx_a:
    output=[cov_all_split[i][j] for j in idx_b if cov_all_split[i][j] is not None]
    if len(output)>0:
      cov_a_b.append(np.hstack(output))
    else:
      cov_a_b.append(None)
  if len(cov_a_b)>1:    
    output2=[x for x in cov_a_b if x is not None]   
    cov_a_b = np.vstack(output2)
  else:
    cov_a_b = np.vstack(cov_a_b)
  if not inv:
    return cov_a_b
  else:
    return np.linalg.pinv(cov_a_b)

def _get_mean_a_given_b(x_all_split, mean_all_split,
    cov_all_split, sub_a, sub_b):
    '''
    Obtain the conditonal mean a given b by using conditonal Gaussian formula
    '''      

    if 'k' not in sub_b:                    
        x_b = _get_x(x_all_split, sub_b)
        mean_a = _get_mean(mean_all_split, sub_a)
        mean_b = _get_mean(mean_all_split, sub_b)
        if mean_b is not None: # consider the case that data in sub_b does not exist
            sigma_a_b = _get_sigma(cov_all_split, sub_a, sub_b)
            inv_sigma_b_b = _get_sigma(cov_all_split, sub_b, sub_b, inv=True)
            output=mean_a + sigma_a_b.dot(inv_sigma_b_b).dot(x_b - mean_b)
        else:
           output=mean_a
        return output
    else:
        nlim=np.asarray(x_all_split[0]).size
        smtx=np.ones((1,nlim))
        if nlim>1:
            idx = [KHS_DICT[i] for i in sub_b if i is not 'k']
            xhs=np.vstack([x_all_split[i] for i in idx])
            x_b=[x_all_split[0].reshape((1,nlim)),xhs.dot(smtx)]
            x_b=np.vstack(x_b)
        else:            
           x_b = _get_x(x_all_split, sub_b)
        
        mean_a = _get_mean(mean_all_split, sub_a)
        mean_b = _get_mean(mean_all_split, sub_b)
        if mean_b is not None: # consider the case that data in sub_b does not exist
            sigma_a_b = _get_sigma(cov_all_split, sub_a, sub_b)
            inv_sigma_b_b = _get_sigma(cov_all_split, sub_b, sub_b, inv=True)
            output=mean_a + sigma_a_b.dot(inv_sigma_b_b).dot(x_b - mean_b)
        else:
            output=mean_a
        return output

def _get_sigma_a_given_b(cov_all_split, sub_a, sub_b):
  '''
  Obtain the conditional covariance a given b by using conditonal Gaussian 
  formula
  '''
  sigma_a_a = _get_sigma(cov_all_split, sub_a, sub_a)
  sigma_a_b = _get_sigma(cov_all_split, sub_a, sub_b)
  if sigma_a_b.size > 0:
    inv_sigma_b_b = _get_sigma(cov_all_split, sub_b, sub_b, inv=True)
    sigma_b_a = _get_sigma(cov_all_split, sub_b, sub_a)
    return sigma_a_a - sigma_a_b.dot(inv_sigma_b_b).dot(sigma_b_a)
  else:
    return sigma_a_a

def _get_multivariate_normal_pdf(x_all_split, mean_all_split,
    cov_all_split, sub_multi):
    '''
    Obtain multivariate Gaussian pdf or conditional multivariate Gaussian
    based upon the specified notations, i.e., sub_multi
    
    Note:
    sub_multi     string    h, s, and k for hard, soft and estimation locations
                            a_b represents a given b, e.g., k_h
    '''                                   
    if "_" in sub_multi:  # "given" type
        sub_a, sub_b = sub_multi.split('_')#x_sub.split('_') (temporarily change by HL)
        m = _get_mean_a_given_b(x_all_split, mean_all_split,
                                cov_all_split, sub_a, sub_b)
        v = _get_sigma_a_given_b(cov_all_split, sub_a, sub_b)
    else:  # single_sub
        sub_a = sub_multi
        m = _get_mean(mean_all_split, sub_a)
        v = _get_sigma(cov_all_split, sub_a, sub_a)
    return scipy.stats.multivariate_normal(m.T[0], v).pdf

def _get_fs(zs):
    '''
    the product of fs distributions
    x   ndim(e.g. npts) by ns
            x is a 2-D np array with the dimension of 
            ndim(number of samples at each integral) 
            by ns (the number of integrals, i.e., number 
            of soft data)      
    '''
    def fs(x):
        res = np.ones((x.shape[0],1))
        for idx_k, zsi in enumerate(zs):
            pdf_type = get_standard_soft_pdf_type(zsi[0])
            if pdf_type == 2:
                nl = zsi[1][0]
                limi = zsi[2]
                probdens = zsi[3]
                y_i = np.interp(
                    x[:,idx_k:idx_k+1], limi[:nl], probdens[:nl],
                    left = 0., right = 0.)
            elif pdf_type == 1:
                nl = zs[1]
                limi = zs[2]
                probdens = zs[3]
                idd = np.where(x[:, idx_k:idx_k+1] - limi[:nl] > 0)[0]
                y_i = probdens[idd]
            elif pdf_type == 10:
                zm = zsi[1]
                zstd = np.sqrt(zsi[2])
                try:
                    y_i = scipy.stats.norm.pdf(
                        x[:,idx_k:idx_k+1], loc=zm, scale=zstd)
                except FloatingPointError:
                    import pdb
                    pdb.set_trace()
              
            if not (y_i.all() or y_i.any()):
                return np.zeros((x.shape[0], 1))
            else:
                res *= y_i
        return res
    return fs

def _get_Fsinv(zs):

    Fs = pdf2cdf(zs)
    def Fsinv(x):  
        res = np.zeros(x.shape)
        for idx_k, Fsi in enumerate(Fs):
            pdf_type = get_standard_soft_pdf_type(Fsi[0])
            if pdf_type == 2:
                nl = Fsi[1][0]
                limi = Fsi[2]
                probdens = Fsi[3]
                probCDFs = Fsi[4]

                alpha = np.diff(probdens) / np.diff(limi)
                i = np.searchsorted(probCDFs[1:],x[:,idx_k:idx_k+1])
                D = np.abs(probdens[i]**2 + 2*alpha[i] * (x[:,idx_k:idx_k+1] - probCDFs[i]))
                y_i = limi[i] + (-probdens[i] + np.sqrt(D)) / alpha[i]
                
            res[:,idx_k:idx_k+1] = y_i
        return res
    return Fsinv

def _get_int_fg_a_given_b_fs_s(x_all_split, mean_all_split,
  cov_all_split, zs, sub_multi, sub_s='s'):
  '''
  The upper right part and lower right part of the last row of formula (1)
  The evaluation is based upon Eqns. (8) or (9) in the cases of s_h and s_kh 
  respectively
  '''

  #fg='s_kh', fs='s'
  sub_a, sub_b = sub_multi.split('_')
  sigma_a_given_b = _get_sigma_a_given_b(cov_all_split, sub_a, sub_b)
  try:
      inv_sigma_a_given_b = np.linalg.pinv(sigma_a_given_b)
  except np.linalg.LinAlgError as e:
      import pdb
      pdb.set_trace()
      raise e
  # mean_tilde_s = zs[1]  # mean
  # sigma_tilde_s = np.diag(zs[2].T[0])  # cov matrix
  mean_tilde_s = []
  sigma_tilde_s = []
  for zsi in zs:
      mean_tilde_s.append([zsi[1]])
      sigma_tilde_s.append(zsi[2])
  mean_tilde_s = np.array(mean_tilde_s) # mean
  sigma_tilde_s = np.diag(sigma_tilde_s)  # cov matrix
  try:
      inv_sigma_tilde_s = np.linalg.pinv(sigma_tilde_s)
  except np.linalg.LinAlgError as e:
      import pdb
      pdb.set_trace()
      raise e

  sigma_t = np.linalg.pinv(inv_sigma_a_given_b + inv_sigma_tilde_s)
  det_sigma_t = np.linalg.det(sigma_t)
  det_sigma_a_given_b = np.linalg.det(sigma_a_given_b)
  det_sigma_tilde_s = np.linalg.det(sigma_tilde_s)
  ns = mean_tilde_s.shape[0]
  alias_c = (
    np.sqrt(det_sigma_t)
    / np.sqrt(det_sigma_a_given_b * det_sigma_tilde_s)
    )
  fgfs_front = alias_c / (2*np.pi)**(ns/2.)
  alias_c = fgfs_front
  # fgfs_front = np.sqrt(det_sigma_t) /\
  #       ((2*np.pi)**(ns/2.) *
  #        np.sqrt(det_sigma_a_given_b * det_sigma_tilde_s)
  #        )

  mean_a_given_b = _get_mean_a_given_b(
        x_all_split, mean_all_split, cov_all_split, sub_a, sub_b)
  fgfs_1 = np.diag((mean_a_given_b.T).dot(
        inv_sigma_a_given_b).dot(mean_a_given_b))
  fgfs_2 = (mean_tilde_s.T).dot(inv_sigma_tilde_s).dot(mean_tilde_s)
  fgfs_3 = (mean_a_given_b.T).dot(inv_sigma_a_given_b) +\
        (mean_tilde_s.T).dot(inv_sigma_tilde_s)
  fgfs_4 = inv_sigma_tilde_s.dot(mean_tilde_s) +\
        inv_sigma_a_given_b.dot(mean_a_given_b)
  alias_sigma_d = sigma_t
  alias_mean_d = alias_sigma_d.dot(fgfs_4)
  # 1 x 1 array (scalar)
  alias_fgfs1234 = np.asscalar(
      (fgfs_1 + fgfs_2 - np.diag(fgfs_3.dot(sigma_t).dot(fgfs_4)))
      )
  fgfs_end = np.exp((-1/2.) * alias_fgfs1234)
  alias_fgfs1234 = fgfs_end
  # fgfs_end = np.exp(
  #       (-1/2.) * (fgfs_1 + fgfs_2 - np.diag(fgfs_3.dot(sigma_t).dot(fgfs_4))))

  # will check later
  # if fgfs_front * fgfs_end == 0:
  #       import pdb
  #       pdb.set_trace()

  return fgfs_front * fgfs_end, \
        (sigma_t, inv_sigma_a_given_b, inv_sigma_tilde_s, mean_tilde_s,
            alias_c, alias_fgfs1234, alias_mean_d, alias_sigma_d)

def _changetimeform(ck,ch=None,cs=None):
  '''
  Change the time format into float while it is in datetime format
  '''  
  
  if type(ck[0,-1])==np.datetime64:
    origin=ck[0,-1]
    ck[:,-1]=np.double(np.asarray(ck[:,-1],dtype='datetime64')-origin)
    ck=ck.astype(np.double)
    if ch is not None and ch.size>0:
      if (not type(ch[0,-1]==np.datetime64)):
        print('Time format of ch is not consistent with ck (np.datetime64)')
        raise
      ch[:,-1]=np.double(np.asarray(ch[:,-1],dtype='datetime64')-origin)
      ch=ch.astype(np.double)
    if cs is not None and cs.size>0:
      if (not type(cs[0,-1]==np.datetime64)):
        print('Time format of cs is not consistent with ck (np.datetime64)')
        raise
      cs[:,-1]=np.double(np.asarray(cs[:,-1],dtype='datetime64')-origin)
      cs=cs.astype(np.double)
      
  return ck,ch,cs

def _set_nh_ns(ck,ch,cs,nhmax,nsmax,dmax):
  '''
  Set the size of nhmax and nsmax that limits the size of matrix to be allocated
  it can be important for an efficient S/T estimation
  '''
  if dmax is not None and np.all(dmax):
    nhmax = int(nhmax)
    nsmax = int(nsmax)
    dmax = np.array(dmax,ndmin=2)
    return nhmax,nsmax,dmax
  
  if ck[0,:].size<3:
    if dmax is None:
      dmax_=0
      if ch is not None:
        ch=np.array(ch,ndmin=2)
        maxd_h=pdist(ch).max()
        dmax_=np.max([dmax_,maxd_h])
      if cs is not None:
        cs=np.array(cs,ndmin=2)
        maxd_s=pdist(cs).max()
        dmax_=np.max([dmax_,maxd_s])
      dmax=np.array(dmax_).reshape(1,1)

    if nhmax is None:
      if ch is not None:
        nhmax=ch.shape[0]
      else:
        nhmax=0

    if nsmax is None:
      if cs is not None:
        nsmax=cs.shape[0]
      else:
        nsmax=0

    
  else:
    maxd=0
    maxt=0
    if dmax is None:
      if ch is not None:
        dummy=np.random.rand(ch.shape[0],1)
        _,cMS_h,tME_h,_=valstv2stg(ch,dummy)
        if nhmax is None:
          nhmax=cMS_h.shape[0]*3
        maxd_h=pdist(cMS_h).max()
        maxt_h=pdist(tME_h.reshape((tME_h.size,1))).max()
      else:
        maxd_h=0
        maxt_h=0
        nhmax=0
      maxd=np.max([maxd_h,maxd]) 
      maxt=np.max([maxt_h,maxt])
      if cs is not None: 
        dummy=np.random.rand(cs.shape[0],1)
        _,cMS_s,tME_s,_=valstv2stg(cs,dummy)
        if nsmax is None:
          if zs[0]==10 or zs[0] is 'gaussian':
            nsmax=cMS_s.shape[0]*3
          else:
            nsmax=3
        maxd_s=pdist(cMS_s).max()
        maxt_s=pdist(tME_s.reshape((tME_s.size,1))).max()
        maxd=np.max([maxd_s,maxd])
        maxt=np.max([maxt_s,maxt])
      else:
        nsmax=0
        maxd_s=0
        maxt_s=0
      maxd=np.max([maxd_s,maxd])
      maxt=np.max([maxt_s,maxt])
    dmax=np.array([maxd,maxt,np.nan]).reshape(1,3)

  return nhmax,nsmax,dmax

def _stratio(covparam):
  '''
  Estimate the S/T ratio for dmax
  '''
  nm=len(covparam)
  sills= np.array([covparam[k][0] for k in range(nm)])  
  hrange = np.array([covparam[k][1][0] for k in range(nm)]) 
  idx0 = np.where([hrange[k] is not None for k in range(nm)])[0]
  idx=np.where(sills[idx0]==sills.max())[0]
  ratio=covparam[idx0[idx]][1][0]/covparam[idx0[idx]][2][0]
  return ratio

def _bme_posterior_prepare(
    ck, ch=None, cs=None, zh=None, zs=None,
    covmodel=None, covparam=None, covmat=None,
    order=np.nan, options=None,
    nhmax=None, nsmax=None, dmax=None,
    general_knowledge='gaussian',
    #  specific_knowledge='unknown',  
    pdfk=None,pdfh=None,pdfs=None,hk_k=None,hk_h=None,hk_s=None,
    gui_args=None):

    '''
    check and configure arguments and 
        find neighbor ckhs index for bme posterior calculation

    ckhs_idx_list:  [ck_idx, ch_idx, cs_idx] represents
        these ck have the same neighbors ch and cs

    return (output_arguments, configured_arguments):
        a tuple contain arguments
    '''
    print('preparing...', end='')
    if covmat is None:
        if (covmodel is None) or (covparam is None):
            raise ValueError(
                'Covariance model and their associated parameters '\
                'should be specified if no covarinace matrix provided.')


    dk = ck.shape[1]
    nk = ck.shape[0]
    nh = ch.shape[0] if ch is not None else 0
    ns = cs.shape[0] if cs is not None else 0

    ck, ch, cs = _changetimeform(ck, ch, cs)
    nhmax, nsmax, dmax = _set_nh_ns(ck, ch, cs, nhmax, nsmax, dmax)
    if dmax.size == 3 and np.isnan(dmax[0][2]):
        dmax[0][2] = _stratio(covparam)
    stratio = dmax[0][2] if dk == 3 else 1.
      
    if options is None:
        options = BMEoptions()

    if gui_args:
        qpgd = gui_args[0]    
    
    if general_knowledge == 'gaussian':
        order = get_standard_order(order)

        if covmat is not None: #consider covmat if exists, not distance, no need to calculate distance
            ckhs_idx_list = []
            if nh != 0:
                covmat_k_h = covmat[:nk, nk:nk+nh] # slice cov(k x h)
                # sort from big to small, clip with nhmax
                k_by_h_idx = (-covmat_k_h).argsort(axis=1)[:, :nhmax]
                # sort h index (inplace)
                k_by_h_idx.sort(axis=1)
                # make ch_ck_dict
                ch_ck_dict = {}
                for k_idx, h_idx in enumerate(k_by_h_idx):
                    tuple_h_idx = tuple(h_idx)
                    if tuple_h_idx not in ch_ck_dict.keys():
                        k_idx_mul = np.where(
                            np.all(h_idx == k_by_h_idx, axis=1)
                            )[0]
                        ch_ck_dict[tuple_h_idx] = list(k_idx_mul)
                    else:
                        continue #skip duplicated row
                if ns != 0:
                    # slice cov(k x s)
                    covmat_k_s = covmat[:nk, nk+nh:nk+nh+ns]
                    for ch_idx, ck_idx in iteritems(ch_ck_dict):
                        picked_covmat_k_s = covmat_k_s[ck_idx, :]
                        # sort from big to small, clip with nsmax
                        picked_k_by_s_idx =\
                            (-picked_covmat_k_s).argsort(axis=1)[:, :nsmax]
                        # sort s index (inplace)
                        picked_k_by_s_idx.sort(axis=1)
                        # make ch_ck_dict
                        cs_ck_dict = {}
                        for picked_k_idx, s_idx in enumerate(picked_k_by_s_idx):
                            tuple_s_idx = tuple(s_idx)
                            if tuple_s_idx not in cs_ck_dict.keys():
                                picked_k_idx_mul = np.where(
                                    np.all(s_idx == picked_k_by_s_idx, axis=1)
                                    )[0]
                                cs_ck_dict[tuple_s_idx] = list(picked_k_idx_mul)
                            else:
                                continue #skip duplicated row
                        for cs_idx, ck2_idx in iteritems(cs_ck_dict):
                            ck_idx = np.array(ck_idx)
                            ckhs_idx_list.append(
                                [ck_idx[ck2_idx,], ch_idx, cs_idx]
                                )
                else: # ns = 0
                    for ch_idx, ck_idx in iteritems(ch_ck_dict):
                        ckhs_idx_list.append([ck_idx, ch_idx, ()])
            elif nh == 0 and ns != 0: # nh = 0, ns != 0
                covmat_k_s = covmat[:nk, nk+nh:nk+nh+ns] # slice cov(k x s)
                # sort from big to small, clip with nsmax
                k_by_s_idx = (-covmat_k_s).argsort(axis=1)[:, :nsmax]
                # sort s index (inplace)
                k_by_s_idx.sort(axis=1)
                # make cs_ck_dict
                cs_ck_dict = {}
                for k_idx, s_idx in enumerate(k_by_s_idx):
                    tuple_s_idx = tuple(s_idx)
                    if tuple_s_idx not in cs_ck_dict.keys():
                        k_idx_mul = np.where(
                            np.all(s_idx == k_by_s_idx, axis=1)
                            )[0]
                        cs_ck_dict[tuple_s_idx] = list(k_idx_mul)
                    else:
                        continue #skip duplicated row
                for cs_idx, ck_idx in iteritems(cs_ck_dict):
                    ckhs_idx_list.append([ck_idx, (), cs_idx])
            else: # nh = 0, ns = 0
                raise ValueError("nh and ns shouldn't be both 0.")
        else:
            #aggregate ck for same hard data and soft data
            # chs = np.vstack(ch, cs)
            ck_norm = np.copy(ck)
            ck_norm[:, -1] = ck_norm[:, -1] * stratio
            if dk == 3:
                dmax_norm = (dmax[0][0]**2 + (dmax[0][1] * stratio)**2)**0.5
            else:
                dmax_norm = dmax[0][0]

            if isinstance(ch, np.ndarray) and nhmax != 0:
                ch_norm = np.copy(ch)
                ch_norm[:, -1] = ch_norm[:, -1] * stratio
                ch_tree = cKDTree(ch_norm)
            if isinstance(cs, np.ndarray) and nsmax != 0:
                cs_norm = np.copy(cs)
                cs_norm[:, -1] = cs_norm[:, -1] * stratio
                cs_tree = cKDTree(cs_norm)

            ckhs_idx_list = []
            if isinstance(ch, np.ndarray) and nhmax != 0: #has harddata
                ch_ck_dict =\
                    neighbours_index_kd(ck_norm, ch_tree, nhmax, dmax_norm)
                for ch_idx, ck_idx in iteritems(ch_ck_dict):
                    if isinstance(cs, np.ndarray) and nsmax != 0: #both hard and soft
                        picked_ck_norm = ck_norm[ck_idx, :]
                        cs_ck_dict =\
                            neighbours_index_kd(
                                picked_ck_norm, cs_tree, nsmax, dmax_norm
                                )
                        for cs_idx, ck2_idx in iteritems(cs_ck_dict):
                            ck_idx = np.array(ck_idx)
                            ckhs_idx_list.append(
                                [ck_idx[ck2_idx,], ch_idx, cs_idx]
                                )
                    else: #only harddata
                        ckhs_idx_list.append([ck_idx, ch_idx, ()])
            elif isinstance(cs, np.ndarray) and nsmax != 0: #only softdata
                cs_ck_dict =\
                    neighbours_index_kd(ck_norm, cs_tree, nsmax, dmax_norm)
                for cs_idx, ck_idx in iteritems(cs_ck_dict):
                    ckhs_idx_list.append([ck_idx, (), cs_idx])
    else:
        raise ValueError("Now we can not consider non-gaussian GK." )

    configured_arguments =\
        (ck, ch, cs, zh, zs,
        covmodel, covparam, covmat,
        order, options,
        nhmax, nsmax, dmax,
        general_knowledge,
        pdfk, pdfh, pdfs, hk_k, hk_h, hk_s,
        gui_args)
    output_arguments = \
        (ckhs_idx_list,)
    print('done')
    return (output_arguments, configured_arguments)

def __get_khs_size(k,h,s):
    if k.shape[0] == 0:
        raise ValueError('ck can not be empty.')
    else:
        return map(__get_coord_size,[k, h, s])

def __get_coord_size(c):
    n = c.shape[0] if c is not None else 0
    return n

def __get_all_split(ck, ch, cs, zh, zs, covmodel, covparam, covmat, order):
    nk, nh, ns = __get_khs_size(ck, ch, cs)
    x_all_split = _get_x_all_split(nk, zh, zs)
    mean_all_split = _get_mean_all_split(x_all_split, order)
    if covmat is not None:
        cov_all_split = np.vsplit(
            covmat, [nk, nk+nh, nk+nh+ns]
            )[:-1] #exclude final empty array
        cov_all_split = \
            [np.hsplit(c, [nk, nk+nh, nk+nh+ns][:-1])\
            for c in cov_all_split]
    else:
        cov_all_split =\
            _get_cov_all_split(ck, ch, cs, covmodel, covparam)
        covmat = coord2Kcombine(cov_all_split)
    return x_all_split, mean_all_split, cov_all_split, covmat

def __get_normalized_constant(
    x_all_split, mean_all_split, cov_all_split, zs, 
    sub_multi = 's_h', sub_s='s'):

    NC, useful_args = _get_int_fg_a_given_b_fs_s(
        x_all_split, mean_all_split,
        cov_all_split, zs, 
        sub_multi = 's_h', sub_s='s'
        )
    return NC

def __get_covmat_remove_index(covmat):
    if np.linalg.matrix_rank(covmat) < covmat.shape[0]:
        # print('warning: matrix rank less than covmat row')
        has_remove_index = True
        diff_n = covmat.shape[0] - np.linalg.matrix_rank(covmat)
        corrmat = np.corrcoef(covmat)
        remove_index = np.vstack(
            np.unravel_index(
                (np.abs(corrmat).ravel()).argsort(),
                covmat.shape
                )
            ).T
        remove_index = remove_index[
            remove_index[:, 0] < remove_index[:, 1]
            ]
        remove_index_res = np.unique(remove_index[:, 1][-diff_n:])
        remove_index_len = remove_index_res.size
        i = 1
        while remove_index_len < diff_n:
            remove_index_res = np.unique(remove_index[:, 1][-(diff_n+i):])
            remove_index_len = remove_index_res.size
            i += 1
        # print('warning: remove_index_res:', remove_index_res)
    else:
        has_remove_index = False
        remove_index_res = np.array([])
    return has_remove_index, remove_index_res

def __check_normalized_constant(NC, options):
    if NC == 0:
        print('Warning NC is equals to zero.')
        if options['debug']:
            import pdb
            pdb.set_trace()
    elif np.isnan(NC):
        print('NC is equals to NaN.')
        if options['debug']:
            import pdb
            pdb.set_trace()
    elif np.isinf(NC):
        print('NC is equals to Inf.')
        if options['debug']:
            import pdb
            pdb.set_trace()
    return NC

def BMEPosteriorMoments(
    ck, ch=None, cs=None, zh=None, zs=None,
    covmodel=None, covparam=None, covmat=None,
    order=np.nan, options=None,
    nhmax=None, nsmax=None, dmax=None,
    general_knowledge='gaussian',
    #  specific_knowledge='unknown',  
    pdfk=None,pdfh=None,pdfs=None,hk_k=None,hk_h=None,hk_s=None,
    gui_args=None, ck_cov_output=False):
    '''
    ck: n by d numpy 2d array
        the estimated data coordinate, usually, d = 3 for spatial-temporal
        coordinate, first two column for spatial, e.g. x, y, and last column
        for temporal, e.g. t.

    ch: n by d numpy 2d array
        the hard data coordinate.

    cs: n by d numpy 2d array
        the soft data coordinate.

    zh: n by 1 numpy 2d array
        the hard data measurement.

    zs: a sequence of soft data record, e.g. (zs1, zs2, zs3, ..., zsn)
        each zsi(i=1~n) is a sequence of data arguments,
        first item should be softpdftype to determind
        the other rest arguments format, e.g.
        syntax: (softdata_type, *softdata_args)
            zs1 = (1, nl, limi, probdens)
            zs2 = (10, zm, zstd)
        if element is emtpy, put None
            e.g. (zs1, zs2, None, ..., zsn)

    covmodel: 

    covmat:
        covariance matrix, a np 2d array
        with shape (nk+nh+ns) by (nk+nh+ns)
    if covmat provieded, covmodel and covparam are simply skipped.
    
    #SI = integrate (fg_s_given_kh * fs_s) dx_s
    #NC = integrate (fg_s_given_h * fs_s) dx_s
    #pdf_k = (fg_kh * SI) / (fg_h * NC)  # eq.1
    #exp_k = ... # eq.2
    #exp_kp = ... # eq.3
    #if general_knowledge == gaussian and specific_knowledge == unknown
    #exp_k = ... # eq.4
    #var_k = ... # eq.5

    gui_args: a tuple with gui arguments

    
    return 
    '''
    (output_arguments, configured_arguments) = _bme_posterior_prepare(
        ck, ch, cs, zh, zs,
        covmodel, covparam, covmat,
        order, options,
        nhmax, nsmax, dmax,
        general_knowledge,
        pdfk, pdfh, pdfs, hk_k, hk_h, hk_s,
        gui_args)
    (ckhs_idx_list,) = output_arguments

    (ck, ch, cs, zh, zs,
        covmodel, covparam, covmat,
        order, options,
        nhmax, nsmax, dmax,
        general_knowledge,
        pdfk, pdfh, pdfs, hk_k, hk_h, hk_s,
        gui_args) = configured_arguments

    if isinstance(order, np.ndarray):
        has_user_defined_general_knowledge = True
    elif order == 0 or np.isnan(order):
        has_user_defined_general_knowledge = False
    else:
        raise ValueError('order type error')

    nk = ck.shape[0]
    nh = ch.shape[0] if ch is not None else 0
    ns = cs.shape[0] if cs is not None else 0
    zk = np.empty((nk,3)) #to 3rd moments
    cur_cnt = 0
    cum_cnt = 0
    
    if general_knowledge == 'gaussian':
        for ck_idx, ch_idx, cs_idx in ckhs_idx_list:
            ck_idx = np.array(ck_idx, dtype=int)
            ch_idx = np.array(ch_idx, dtype=int)
            cs_idx = np.array(cs_idx, dtype=int)
            picked_ch =\
                ch[ch_idx, :] if isinstance(ch, np.ndarray) else None
            picked_cs =\
                cs[cs_idx, :] if isinstance(cs, np.ndarray) else None
            picked_zh =\
                zh[ch_idx, :] if isinstance(zh, np.ndarray) else None
            picked_zs =\
                [zs[cs_idx_i] for cs_idx_i in cs_idx] if zs is not None else None
            ck_count = ck_idx.shape[0]
            split_count = np.ceil(ck_count/250.)
            for ck_idx_piece in np.array_split(ck_idx, split_count):
                picked_ck = ck[ck_idx_piece, :]
                if covmat is not None:
                    covidx = np.hstack(
                        (ck_idx_piece, nk+ch_idx, nk+nh+cs_idx)
                        )
                    picked_covmat = covmat[np.ix_(covidx, covidx)]
                else:
                    picked_covmat = covmat

                #picked order for specified general knowledge
                if has_user_defined_general_knowledge:
                    order_idx = np.hstack(
                        (ck_idx_piece, nk+ch_idx, nk+nh+cs_idx)
                        )
                    picked_order = order[order_idx, :]
                else:
                    picked_order = order

                try:
                    picked_mvs = _bme_posterior_moments(
                        picked_ck, picked_ch, picked_cs,
                        picked_zh, picked_zs,
                        covmodel, covparam, picked_covmat,
                        picked_order, options, general_knowledge,
                        pdfk, pdfh, pdfs, hk_k, hk_h, hk_s,
                        gui_args, ck_cov_output)
                except np.linalg.LinAlgError as e:
                    import pdb
                    pdb.set_trace()
                    raise e

                
                zk[ck_idx_piece, :] = picked_mvs
                if gui_args:
                    qpgd = gui_args[0]
                    if qpgd.wasCanceled(): #cancel by user
                        return False
                    else:
                        qpgd.setValue(qpgd.value()+ck_idx_piece.size)
                else:
                    cur_cnt += ck_idx_piece.size
                    if cur_cnt - cum_cnt >= 2000:
                        print(cur_cnt, '/', nk)
                        cum_cnt = cur_cnt
        print(cur_cnt, '/', nk)
        return zk
    else:
      nk=len(pdfk)
      moments=np.empty((nk,3))

      for k in range(nk):
        print('BME MOMENTS:' + str(k+1) + '/' + str(nk))
        
        cklocal=ck[k:k+1,:]
        pdfk_local=[pdfk[k]]
        hk_k_local=[hk_k[k]]
        
        pdf_k=BMEPosteriorPDF(cklocal, ch, cs, zh, zs, covmodel, covparam,
              order, options, nhmax, nsmax, dmax, general_knowledge,
              pdfk=pdfk_local,pdfh=pdfh,pdfs=pdfs,
              hk_k=hk_k_local,hk_h=hk_h,hk_s=hk_s)[0]
          
        zmin=hk_k[k][0]-6*np.sqrt(hk_k[k][1])
        zmax=hk_k[k][0]+6*np.sqrt(hk_k[k][1])
        
        xxx=np.linspace(zmin,zmax,100)
        aaa=pdf_k(xxx,0)

        maxpts = options[2][0]
        aEps = 0
        rEps = options[3][0]

        from cubature import cubature

        mon1_for_cubature = lambda x_array: x_array[:,0] * pdf_k(x_array[:,0],0)[:,0]
        mon1,mon1_err = cubature(
            func=mon1_for_cubature, ndim=1, fdim=1, xmin=np.array([zmin]),
            xmax=np.array([zmax]), adaptive='h', maxEval = maxpts,
            abserr = 0, relerr = rEps, vectorized = True)
        mon2_for_cubature = lambda x_array: x_array[:,0]**2 * pdf_k(x_array[:,0],0)[:,0]
        mon2,mon2_err = cubature(
            func=mon2_for_cubature,ndim=1, fdim=1, xmin=np.array([zmin]),
            xmax=np.array([zmax]), adaptive='h', maxEval = maxpts,
            abserr = 0, relerr = rEps, vectorized = True)  
        mon3_for_cubature = lambda x_array: x_array[:,0]**3 * pdf_k(x_array[:,0],0)[:,0]
        mon3,mon3_err = cubature(
            func=mon3_for_cubature,ndim=1, fdim=1, xmin=np.array([zmin]),
            xmax=np.array([zmax]), adaptive='h', maxEval = maxpts,
            abserr = 0, relerr = rEps, vectorized = True)

        moments[k,0]=mon1
        moments[k,1]=mon2-mon1**2
        moments[k,2]=mon3-3*mon1*mon2-mon1**3

      return moments[:,0], moments[:,1], moments[:,2]

def BMEPosteriorPDF(
    ck, ch=None, cs=None, zh=None, zs=None,
    covmodel=None, covparam=None, covmat=None,
    order=np.nan, options=None,
    nhmax=None, nsmax=None, dmax=None,
    general_knowledge='gaussian',
    #  specific_knowledge='unknown',  
    pdfk=None,pdfh=None,pdfs=None,hk_k=None,hk_h=None,hk_s=None,
    gui_args=None):

    (output_arguments, configured_arguments) = _bme_posterior_prepare(
        ck, ch, cs, zh, zs,
        covmodel, covparam, covmat,
        order, options,
        nhmax, nsmax, dmax,
        general_knowledge,
        pdfk, pdfh, pdfs, hk_k, hk_h, hk_s,
        gui_args)
    (ckhs_idx_list,) = output_arguments

    (ck, ch, cs, zh, zs,
        covmodel, covparam, covmat,
        order, options,
        nhmax, nsmax, dmax,
        general_knowledge,
        pdfk, pdfh, pdfs, hk_k, hk_h, hk_s,
        gui_args) = configured_arguments
   
    if isinstance(order, np.ndarray):
        has_user_defined_general_knowledge = True
    elif order == 0 or np.isnan(order):
        has_user_defined_general_knowledge = False
    else:
        raise ValueError('order type error')
    nk = ck.shape[0]
    nh = ch.shape[0] if ch is not None else 0
    ns = cs.shape[0] if cs is not None else 0
    zk = np.empty((ck.shape[0],1), dtype=object) # to 1 pdf function
    if general_knowledge == 'gaussian':
        for ck_idx, ch_idx, cs_idx in ckhs_idx_list:
            ck_idx = np.array(ck_idx, dtype=int)
            ch_idx = np.array(ch_idx, dtype=int)
            cs_idx = np.array(cs_idx, dtype=int)
            picked_ch =\
                ch[ch_idx, :] if isinstance(ch, np.ndarray) else None
            picked_cs =\
                cs[cs_idx, :] if isinstance(cs, np.ndarray) else None
            picked_zh =\
                zh[ch_idx, :] if isinstance(zh, np.ndarray) else None
            picked_zs =\
                [zs[cs_idx_i] for cs_idx_i in cs_idx] if zs is not None else None
            ck_count = ck_idx.shape[0]
            split_count = np.ceil(ck_count/250.)
            for ck_idx_piece in np.array_split(ck_idx, split_count):
                picked_ck = ck[ck_idx_piece, :]
                if covmat is not None:
                    covidx = np.hstack(
                        (ck_idx_piece, nk+ch_idx, nk+nh+cs_idx)
                        )
                    picked_covmat = covmat[np.ix_(covidx, covidx)]
                else:
                    picked_covmat = covmat

                #picked order for specified general knowledge
                if has_user_defined_general_knowledge:
                    order_idx = np.hstack(
                        (ck_idx_piece, nk+ch_idx, nk+nh+cs_idx)
                        )
                    picked_order = order[order_idx, :]
                else:
                    picked_order = order

                try:
                    picked_mvs = _bme_posterior_pdf(
                        picked_ck, picked_ch, picked_cs,
                        picked_zh, picked_zs,
                        covmodel, covparam, picked_covmat,
                        picked_order, options, general_knowledge,
                        pdfk, pdfh, pdfs, hk_k, hk_h, hk_s,
                        gui_args)
                except Exception as e:
                    print(e)
                    import pdb
                    pdb.set_trace()
                    raise e
                zk[ck_idx_piece, :] = picked_mvs
                if gui_args:
                    if qpgd.wasCanceled(): #cancel by user
                        return False
                    else:
                        qpgd.setValue(qpgd.value()+ck_idx_piece.size)
        return zk
